"""Public DGS Corpus: parallel corpus for German Sign Language (DGS) with German and English annotations"""
from __future__ import annotations

import gzip
import json
from copy import copy

import cv2
import math

import numpy as np
import tensorflow as tf
import tensorflow_datasets as tfds

from os import path
from typing import Dict, Any, Set, Optional, List

from pose_format.numpy import NumPyPoseBody
from pose_format.utils.openpose import load_openpose, OpenPoseFrames
from pose_format.pose import Pose

from ..warning import dataset_warning
from ...datasets.config import SignDatasetConfig
from ...utils.features import PoseFeature

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

_DESCRIPTION = """
Parallel corpus for German Sign Language (DGS) with German and English annotations
"""

_CITATION = """
@misc{dgscorpus_3,
  title = {MEINE DGS -- annotiert. {\"O}ffentliches Korpus der Deutschen Geb{\"a}rdensprache, 3. Release / MY DGS -- annotated. Public Corpus of German Sign Language, 3rd release},
  author = {Konrad, Reiner and Hanke, Thomas and Langer, Gabriele and Blanck, Dolly and Bleicken, Julian and Hofmann, Ilona and Jeziorski, Olga and K{\"o}nig, Lutz and K{\"o}nig, Susanne and Nishio, Rie and Regen, Anja and Salden, Uta and Wagner, Sven and Worseck, Satu and B{\"o}se, Oliver and Jahn, Elena and Schulder, Marc},
  year = {2020},
  type = {languageresource},
  version = {3.0},
  publisher = {Universit{\"a}t Hamburg},
  url = {https://doi.org/10.25592/dgs.corpus-3.0},
  doi = {10.25592/dgs.corpus-3.0}
}
"""

_HOMEPAGE = "https://www.sign-lang.uni-hamburg.de/meinedgs/"

# This `dgs.json` file was created using `create_index.py`
INDEX_URL = "https://nlp.biu.ac.il/~amit/datasets/dgs.json"

_POSE_HEADERS = {
    "holistic": path.join(path.dirname(path.realpath(__file__)), "holistic.poseheader"),
    "openpose": path.join(path.dirname(path.realpath(__file__)), "openpose.poseheader"),
}

_KNOWN_SPLITS = {
    "3.0.0-uzh-document": path.join(path.dirname(path.realpath(__file__)), "splits", "split.3.0.0-uzh-document.json"),
    "3.0.0-uzh-sentence": path.join(path.dirname(path.realpath(__file__)), "splits", "split.3.0.0-uzh-sentence.json")
}


def convert_dgs_dict_to_openpose_frames(input_dict: Dict[str, Any]) -> OpenPoseFrames:
    """
    Modifies the DGS Openpose format slightly to be compatible with the `pose_format` library. Most notably,
    changes the dict keys to integers (from strings) and adds a "frame_id" key to each frame dict.

    :param input_dict: OpenPose output, one dictionary overall, one sub-dictionary for each frame.
    :return: A dictionary of OpenPose frames that is compatible with `pose_format.utils.load_openpose`.
    """
    frames = {}  # type: OpenPoseFrames

    for frame_id_as_string, frame_dict in input_dict.items():
        frame_id = int(frame_id_as_string)
        frame_dict["frame_id"] = frame_id
        frames[frame_id] = frame_dict

    return frames


def get_openpose(openpose_path: str, fps: int, people: Optional[Set] = None, num_frames: Optional[int] = None) -> Dict[str, Pose]:
    """
    Load OpenPose in the particular format used by DGS (one single file vs. one file for each frame).

    :param openpose_path: Path to a file that contains OpenPose for all frames of a DGS video, in one single JSON.
    :param fps: Framerate.
    :param people: Specify the name prefixes of camera views that should be extracted. The default value is {"a", "b"}.
    :param num_frames: Number of frames when it is known and cannot be derived from OpenPose files. That is the case if
                       the last frame(s) of a video are missing from the OpenPose output.
    :return: Dictionary of Pose objects (one for each person) with a header specific to OpenPose and a body that
             contains a single array.
    """
    # set mutable default argument
    if people is None:
        people = {"a", "b"}

    with gzip.GzipFile(openpose_path, "r") as openpose_raw:
        openpose = json.loads(openpose_raw.read().decode("utf-8"))

    # select views if their name starts with "a" or "b"
    views = {view["camera"][0]: view for view in openpose if view["camera"][0] in people}

    poses = {p: None for p in people}
    for person, view in views.items():
        width, height, frames_obj = view["width"], view["height"], view["frames"]

        # Convert to pose format
        frames = view["frames"]
        frames = convert_dgs_dict_to_openpose_frames(frames)
        poses[person] = load_openpose(frames, fps=fps, width=width, height=height, depth=0, num_frames=num_frames)

    return poses


def load_split(split_name: str) -> Dict[str, List[str]]:
    """
    Loads a split from the file system. What is loaded must be a JSON object with the following structure:

    {"train": ..., "dev": ..., "test": ...}

    :param split_name: An identifier for a predefined split or a filepath to a custom split file.
    :return: The split loaded as a dictionary.
    """
    if split_name not in _KNOWN_SPLITS.keys():
        # assume that the supplied string is a path on the file system
        if not path.exists(split_name):
            raise ValueError(
                "Split '%s' is not a known data split identifier and does not exist as a file either.\n"
                "Known split identifiers are: %s" % (split_name, str(_KNOWN_SPLITS))
            )

        split_path = split_name
    else:
        # the supplied string is an identifier for a predefined split
        split_path = _KNOWN_SPLITS[split_name]

    with open(split_path) as infile:
        split = json.load(infile)  # type: Dict[str, List[str]]

    return split


DEFAULT_FPS = 50


class DgsCorpusConfig(SignDatasetConfig):
    def __init__(self, data_type: Literal["document", "sentence"] = "document", split: str = None, **kwargs):
        """
        :param split: An identifier for a predefined split or a filepath to a custom split file.
        :param data_type: Whether to return documents or sentences as data.
        """
        super().__init__(**kwargs)

        self.data_type = data_type
        self.split = split

        # Verify split matches data type
        if self.split in _KNOWN_SPLITS and not self.split.endswith(self.data_type):
            raise ValueError(f"Split '{self.split}' is not compatible with data type '{self.data_type}'.")


class DgsCorpus(tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for dgs_corpus dataset."""

    VERSION = tfds.core.Version("3.0.0")
    RELEASE_NOTES = {
        "3.0.0": "3rd release",
    }

    BUILDER_CONFIGS = [
        DgsCorpusConfig(name="default", include_video=True, include_pose="openpose"),
        DgsCorpusConfig(name="videos", include_video=True, include_pose=None),
        DgsCorpusConfig(name="openpose", include_video=False, include_pose="openpose"),
        DgsCorpusConfig(name="holistic", include_video=False, include_pose="holistic"),
        DgsCorpusConfig(name="annotations", include_video=False, include_pose=None),
        DgsCorpusConfig(name="sentences", include_video=False, include_pose=None, data_type="sentence"),
    ]

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""

        assert isinstance(self._builder_config, DgsCorpusConfig), \
            "Builder config for dgs_corpus must be an instance of DgsCorpusConfig"

        features = {
            "id": tfds.features.Text(),
            "paths": {
                "ilex": tfds.features.Text(),
                "eaf": tfds.features.Text(),
                "srt": tfds.features.Text(),
                "cmdi": tfds.features.Text(),
            },
        }

        if self._builder_config.data_type == "sentence":
            features["document_id"] = tfds.features.Text()
            features["sentence"] = {
                "id": tfds.features.Text(),
                "participant": tfds.features.Text(),
                "start": tf.int32,
                "end": tf.int32,
                "german": tfds.features.Text(),
                "english": tfds.features.Text(),
                "glosses": tfds.features.Sequence(
                    {
                        "start": tf.int32,
                        "end": tf.int32,
                        "gloss": tfds.features.Text(),
                        "hand": tfds.features.Text(),
                        "Lexeme_Sign": tfds.features.Text(),
                        "Gebärde": tfds.features.Text(),
                        "Sign": tfds.features.Text(),
                    }
                ),
                "mouthings": tfds.features.Sequence({"start": tf.int32, "end": tf.int32, "mouthing": tfds.features.Text()}),
            }

        video_ids = ["a", "b", "c"]

        if self._builder_config.include_video:
            features["fps"] = tf.int32
            if self._builder_config.process_video:
                video_feature = self._builder_config.video_feature((640, 360))
                if self._builder_config.data_type == "document":
                    features["videos"] = {_id: video_feature for _id in video_ids}
                else:
                    features["video"] = video_feature
            features["paths"]["videos"] = {_id: tfds.features.Text() for _id in video_ids}

        # Add poses if requested
        if self._builder_config.include_pose is not None:
            pose_header_path = _POSE_HEADERS[self._builder_config.include_pose]
            stride = 1 if self._builder_config.fps is None else 50 / self._builder_config.fps

            if self._builder_config.include_pose == "openpose":
                pose_shape = (None, 1, 137, 2)
            if self._builder_config.include_pose == "holistic":
                pose_shape = (None, 1, 543, 3)

            pose_feature = PoseFeature(shape=pose_shape, stride=stride, header_path=pose_header_path)
            if self._builder_config.data_type == "document":
                features["poses"] = {_id: pose_feature for _id in ["a", "b"]}
            else:
                features["pose"] = pose_feature

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features),
            homepage=_HOMEPAGE,
            supervised_keys=None,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        """Returns SplitGenerators."""
        dataset_warning(self)

        index_path = dl_manager.download(INDEX_URL)

        with open(index_path, "r", encoding="utf-8") as f:
            index_data = json.load(f)

        # No need to download HTML pages
        for datum in index_data.values():
            del datum["transcript"]
            del datum["format"]

        # Don't download videos if not necessary
        if not self._builder_config.include_video:
            for datum in index_data.values():
                del datum["video_a"]
                del datum["video_b"]
                del datum["video_c"]

        # Don't download openpose poses if not necessary
        if self._builder_config.include_pose != "openpose":
            for datum in index_data.values():
                del datum["openpose"]

        # Don't download holistic poses if not necessary
        if self._builder_config.include_pose != "holistic":
            for datum in index_data.values():
                del datum["holistic_a"]
                del datum["holistic_b"]

        urls = {url: url for datum in index_data.values() for url in datum.values() if url is not None}

        local_paths = dl_manager.download(urls)

        data = {_id: {k: local_paths[v] if v is not None else None for k, v in datum.items()} for _id, datum in index_data.items()}

        if self._builder_config.split is not None:
            split = load_split(self._builder_config.split)

            train_args = {"data": data, "split": split["train"]}
            dev_args = {"data": data, "split": split["dev"]}
            test_args = {"data": data, "split": split["test"]}

            return [
                tfds.core.SplitGenerator(name=tfds.Split.TRAIN, gen_kwargs=train_args),
                tfds.core.SplitGenerator(name=tfds.Split.VALIDATION, gen_kwargs=dev_args),
                tfds.core.SplitGenerator(name=tfds.Split.TEST, gen_kwargs=test_args),
            ]

        else:
            return [tfds.core.SplitGenerator(name=tfds.Split.TRAIN, gen_kwargs={"data": data})]

    def _include_videos(self, datum: Any, features: Any):
        videos = {}
        for participant in ["a", "b", "c"]:
            video_key = "video_" + participant

            if video_key not in datum.keys() or datum[video_key] is None:
                video = ""
            else:
                video = str(datum[video_key])

            videos[participant] = video

        # make sure that the video fps is as expected
        for video_path in videos.values():
            if video_path == "":
                continue

            cap = cv2.VideoCapture(video_path)
            actual_video_fps = cap.get(cv2.CAP_PROP_FPS)
            cap.release()

            assert math.isclose(actual_video_fps, float(DEFAULT_FPS)), "Framerate of video '%s' is %f instead of %d" % (
                video_path,
                actual_video_fps,
                DEFAULT_FPS,
            )

        features["fps"] = self._builder_config.fps if self._builder_config.fps is not None else DEFAULT_FPS
        features["paths"]["videos"] = videos

    def _generate_examples(self, data, split: List[str] | Dict[str, List[str]] = None):
        """ Yields examples. """
        from .dgs_utils import get_elan_sentences

        default_video = np.zeros((0, 0, 0, 3))  # Empty video

        for document_id, datum in list(data.items()):
            if split is not None and document_id not in split:
                continue

            features = {
                "id": document_id,
                "paths": {t: str(datum[t]) if t in datum else "" for t in ["ilex", "eaf", "srt", "cmdi"]},
            }

            if self._builder_config.include_video:
                self._include_videos(datum, features)

            poses = None
            if self._builder_config.include_pose is not None:
                if self._builder_config.include_pose == "openpose":
                    poses = get_openpose(datum["openpose"], fps=DEFAULT_FPS)

                if self._builder_config.include_pose == "holistic":
                    poses = {}
                    for person in ["a", "b"]:
                        if datum["holistic_" + person] is not None:
                            with open(datum["holistic_" + person], "rb") as f:
                                poses[person] = Pose.read(f.read())
                        else:
                            poses[person] = None

            if self._builder_config.data_type == "document":
                if poses is not None:
                    features["poses"] = poses
                if self._builder_config.process_video:
                    features["videos"] = {t: v if v != "" else default_video for t, v in features["paths"]["videos"].items()}
                features["id"] = document_id
                yield document_id, features
            else:
                sentences = list(get_elan_sentences(datum["eaf"]))
                for sentence in sentences:
                    if split is not None and sentence["id"] not in split[document_id]:
                        continue

                    if sentence["english"] is None:
                        sentence["english"] = ""

                    features = copy(features)  # Unclear if necessary, but better safe than sorry
                    features["sentence"] = sentence

                    start_time = sentence["start"] / 1000
                    start_frame = int(start_time * DEFAULT_FPS)
                    end_time = sentence["end"] / 1000
                    end_frame = math.ceil(end_time * DEFAULT_FPS)

                    if poses is not None:
                        pose = poses[sentence["participant"].lower()]
                        sub_pose_body = NumPyPoseBody(
                            fps=pose.body.fps,
                            data=pose.body.data[start_frame:end_frame+1],
                            confidence=pose.body.confidence[start_frame:end_frame+1],
                        )
                        features["pose"] = Pose(pose.header, sub_pose_body)

                    if self._builder_config.process_video:
                        videos = features["paths"]["videos"]
                        if videos[sentence["participant"].lower()] == "":
                            features["video"] = default_video
                        else:
                            features["video"] = {
                                "video": videos[sentence["participant"].lower()],
                                "ffmpeg_args": ["-ss", str(start_time), "-to", str(end_time)],
                            }
                    document_sentence_id = f'{document_id}_{sentence["id"]}'
                    features["document_id"] = document_id
                    features["id"] = document_sentence_id
                    yield document_sentence_id, features



import warnings


def dataset_warning(dataset):
    warnings.warn(
        "This library provides access to data sets without claiming ownership over them or defining their licensing terms. "
        "Users who download data are responsible for checking the license of each individual data set."
    )



"""rwth_phoenix_2014_t dataset."""

import tensorflow_datasets as tfds

from . import rwth_phoenix2014_t


class RWTHPhoenix2014TVideosTest(tfds.testing.DatasetBuilderTestCase):
    """Tests for rwth_phoenix_2014_t dataset."""

    DATASET_CLASS = rwth_phoenix2014_t.RWTHPhoenix2014T
    SPLITS = {"train": 1, "validation": 1, "test": 1}
    OVERLAPPING_SPLITS = ["train", "validation", "test"]
    BUILDER_CONFIG_NAMES_TO_TEST = ["videos"]

    DL_EXTRACT_RESULT = ["annotations"]


class RWTHPhoenix2014TPosesTest(tfds.testing.DatasetBuilderTestCase):
    """Tests for rwth_phoenix_2014_t dataset."""

    DATASET_CLASS = rwth_phoenix2014_t.RWTHPhoenix2014T
    SPLITS = {"train": 1, "validation": 1, "test": 1}
    OVERLAPPING_SPLITS = ["train", "validation", "test"]
    BUILDER_CONFIG_NAMES_TO_TEST = ["poses"]

    DL_EXTRACT_RESULT = ["annotations", "poses"]


if __name__ == "__main__":
    tfds.testing.test_main()



"""Helper file to crawl The Public DGS Corpus and create an up-to-date index of the dataset"""

import json
import re
import urllib.request

corpus_path = "https://www.sign-lang.uni-hamburg.de/meinedgs/"

index_data = {}

with urllib.request.urlopen(corpus_path + "ling/start-name_en.html") as response:
    html = response.read().decode("utf-8")

    trs = re.findall('<tr id="(.*?)">([\s\S]*?)</tr>', html)
    for tr_id, tr in trs:
        tds = re.findall("(?:(<td />)|<td.*?>([\s\S]*?)</)", tr)
        td_links = [re.findall('href="(.*?)"', td[1]) for td in tds]
        links = [corpus_path + links[0][3:] if len(links) > 0 else None for links in td_links]
        assert len(links) == 13

        index_data[tr_id] = {
            "transcript": links[0],
            "format": links[2],
            "ilex": links[4],
            "eaf": links[5],
            "video_a": links[6],
            "video_b": links[7],
            "video_c": links[8],
            "srt": links[9],
            "cmdi": links[11],
            "openpose": links[12],
        }

        # Add holistic
        for c in ["a", "b"]:
            holistic_path = "https://nlp.biu.ac.il/~amit/datasets/poses/holistic/dgs_corpus/" + tr_id + "_" + c + ".pose"
            index_data[tr_id]["holistic_" + c] = holistic_path if index_data[tr_id]["video_" + c] is not None else None

        # Make sure parsing worked
        if index_data[tr_id]["openpose"] is not None:
            assert index_data[tr_id]["openpose"].endswith(".json.gz")

with open("data.json", "w") as f:
    json.dump(index_data, f)



"""dgs_corpus dataset."""

import json
import gzip
import shutil
import tempfile
import numpy as np
import tensorflow_datasets as tfds

from unittest import TestCase
from typing import List, Dict
from contextlib import contextmanager

from sign_language_datasets.datasets.dgs_corpus import dgs_corpus
from pose_format.pose import Pose


OPENPOSE_TOTAL_KEYPOINTS = 137
OPENPOSE_COMPONENTS_USED = ["pose_keypoints_2d", "face_keypoints_2d", "hand_left_keypoints_2d", "hand_right_keypoints_2d"]
OPENPOSE_COMPONENTS_UNUSED = ["pose_keypoints_3d", "face_keypoints_3d", "hand_left_keypoints_3d", "hand_right_keypoints_3d"]
OPENPOSE_NUM_POINTS_PER_COMPONENT = {
    "pose_keypoints_2d": 25,
    "face_keypoints_2d": 70,
    "hand_left_keypoints_2d": 21,
    "hand_right_keypoints_2d": 21,
}


def _generate_dgs_openpose_data(camera_names: List[str], num_frames: int = 10, num_people: int = 1) -> List[Dict]:
    data = []  # type: List[Dict]

    for camera_name in camera_names:
        camera_dict = {"camera": camera_name}
        camera_dict["id"] = np.random.randint(10000000)
        camera_dict["width"] = 1280
        camera_dict["height"] = 720

        camera_dict["frames"] = {}

        for frame_id in range(num_frames):
            frame_dict = {"version": 1.4, "people": []}

            for _ in range(num_people):
                keypoints_dict = {}
                for component in OPENPOSE_COMPONENTS_USED:
                    num_keypoints = OPENPOSE_NUM_POINTS_PER_COMPONENT[component]

                    # for each keypoint: one value for each dimension, plus confidence
                    num_pose_values = 3 * num_keypoints
                    pose_data = np.random.random_sample((num_pose_values,)).tolist()
                    keypoints_dict[component] = pose_data

                for component in OPENPOSE_COMPONENTS_UNUSED:
                    keypoints_dict[component] = []

                frame_dict["people"].append(keypoints_dict)

            camera_dict["frames"][str(frame_id)] = frame_dict

        data.append(camera_dict)

    return data


def _generate_dgs_openpose_file(
    filehandle: tempfile.NamedTemporaryFile, camera_names: List[str], num_frames: int = 10, num_people: int = 1
) -> None:

    data = _generate_dgs_openpose_data(camera_names, num_frames, num_people)
    json.dump(data, filehandle)
    filehandle.flush()


def _gzip_file(filepath_in: str, filepath_out: str) -> None:
    with open(filepath_in, "rb") as filehandle_in:
        with gzip.open(filepath_out, "wb") as filehandle_out:
            shutil.copyfileobj(filehandle_in, filehandle_out)


@contextmanager
def _create_tmp_dgs_openpose_file(camera_names: List[str], num_frames: int = 10, num_people: int = 1) -> str:
    with tempfile.NamedTemporaryFile(mode="w+") as filehandle:
        _generate_dgs_openpose_file(filehandle, camera_names, num_frames, num_people)

        filepath_zipped = filehandle.name + ".gz"
        _gzip_file(filehandle.name, filepath_zipped)

        yield filepath_zipped


class TestDgsCorpusAuxiliaryFunctions(TestCase):
    def test_convert_dgs_dict_to_openpose_frames(self):

        input_dict = {"7": {"people": [1, 2, 3]}, "8": {"people": [4, 5, 6]}}

        expected_output = {7: {"people": [1, 2, 3], "frame_id": 7}, 8: {"people": [4, 5, 6], "frame_id": 8}}

        actual_output = dgs_corpus.convert_dgs_dict_to_openpose_frames(input_dict)

        self.assertDictEqual(actual_output, expected_output)

    def test_get_poses_return_type(self):

        camera_names_in_mock_data = ["a", "b", "c"]
        num_frames_in_mock_data = 10
        num_people_in_mock_data = 1

        people_to_extract = {"a", "b"}

        with _create_tmp_dgs_openpose_file(
            camera_names=camera_names_in_mock_data, num_frames=num_frames_in_mock_data, num_people=num_people_in_mock_data
        ) as filepath:
            poses = dgs_corpus.get_openpose(openpose_path=filepath, fps=50, people=people_to_extract)

            for pose in poses.values():
                self.assertIsInstance(pose, Pose)

    def test_get_poses_subset_of_camera_names(self):

        camera_names_in_mock_data = ["a2", "b1", "c5"]
        num_frames_in_mock_data = 10
        num_people_in_mock_data = 1

        people_to_extract = {"a", "b"}

        with _create_tmp_dgs_openpose_file(
            camera_names=camera_names_in_mock_data, num_frames=num_frames_in_mock_data, num_people=num_people_in_mock_data
        ) as filepath:
            poses = dgs_corpus.get_openpose(openpose_path=filepath, fps=50, people=people_to_extract)

            for person in poses.keys():
                self.assertTrue(person[0] in people_to_extract)


'''
class DgsCorpusTest(tfds.testing.DatasetBuilderTestCase):
    """Tests for dgs_corpus dataset."""

    # TODO(dgs_corpus):
    DATASET_CLASS = dgs_corpus.DgsCorpus
    SPLITS = {
        "train": 3,  # Number of fake train example
        "test": 1,  # Number of fake test example
    }

    # If you are calling `download/download_and_extract` with a dict, like:
    #   dl_manager.download({'some_key': 'http://a.org/out.txt', ...})
    # then the tests needs to provide the fake output paths relative to the
    # fake data directory
    # DL_EXTRACT_RESULT = {'some_key': 'output_file1.txt', ...}
'''

if __name__ == "__main__":
    tfds.testing.test_main()



import json

import tensorflow_datasets as tfds
import hashlib

# noinspection PyUnresolvedReferences
import sign_language_datasets.datasets
from sign_language_datasets.datasets.config import SignDatasetConfig

def sign2mint():
    config = SignDatasetConfig(name="firestore", version="1.0.0", include_video=False)
    dataset = tfds.load(name='sign2_mint', builder_kwargs={"config": config})

    for datum in dataset["train"]:
        name = datum['fachbegriff'].numpy().decode('utf-8')
        swu = datum['swu'].numpy().decode('utf-8')
        url = datum['video'].numpy().decode('utf-8')
        uid_raw = "sign2mint" + datum['id'].numpy().decode('utf-8')
        uid = 's2m' + hashlib.md5(uid_raw.encode()).hexdigest()

        doc = {
            "uid": uid,
            "url": url,
            "meta": {
                "visibility": "unlisted",
                "name": "Sign2MINT: " + name,
                "language": "gsg",
                "userId": "1aGEJfRxDqdpIbOnjU5i5X0quYu2"
            },
            "context": {
                "id": uid_raw,
                "wiktionary": datum['wortlink'].numpy().decode('utf-8'),
                "definition": datum['definition'].numpy().decode('utf-8')
            }
        }

        captions = [
            {"language": "de", "transcription": name},
            {"language": "Sgnw", "transcription": swu},
        ]

        yield doc, captions, None


if __name__ == '__main__':
    with open('sign2mint.jsonl', 'w') as f:
        for doc, captions, _ in sign2mint():
            f.write(json.dumps({"doc": doc, "captions": captions}))
            f.write('\n')



import itertools

import tensorflow_datasets as tfds
from dotenv import load_dotenv

# noinspection PyUnresolvedReferences
# from sign_language_datasets.datasets.dgs_corpus import DgsCorpusConfig

import sign_language_datasets.datasets

# noinspection PyUnresolvedReferences
# import sign_language_datasets.datasets.dgs_corpus
from sign_language_datasets.datasets.config import SignDatasetConfig
# from sign_language_datasets.datasets.dgs_corpus.dgs_corpus import DgsCorpusConfig

load_dotenv()

# config = SignDatasetConfig(name="only-annotations", version="3.0.0", include_video=False)
# rwth_phoenix2014_t = tfds.load(name="rwth_phoenix2014_t", builder_kwargs=dict(config=config))

# config = SignDatasetConfig(name="256x256:10", include_video=True, fps=10, resolution=(256, 256))

# aslg_pc12 = tfds.load('aslg_pc12')
#
# rwth_phoenix2014_t = tfds.load('rwth_phoenix2014_t', builder_kwargs=dict(config=config))

# wlasl = tfds.load('wlasl', builder_kwargs=dict(config=config))
#
# autsl = tfds.load('autsl', builder_kwargs=dict(
#     config=SignDatasetConfig(name="test", include_video=False, include_pose="holistic"),
# ))

# dgs_config = DgsCorpusConfig(name="sentence-test-video", data_type="sentence",
#                              include_video=False, process_video=False, include_pose=None)
# dgs_corpus = tfds.load('dgs_corpus', builder_kwargs=dict(config=dgs_config))
#
# for datum in itertools.islice(dgs_corpus["train"], 0, 10):
#   print(datum)


# config = DgsCorpusConfig(name="holistic-features", version="3.0.0", include_video=False,
#                          include_pose="holistic", split="3.0.0-uzh-document")
# dgs_corpus = tfds.load('dgs_corpus', builder_kwargs=dict(config=config))

config = SignDatasetConfig(name="signsuisse", version="1.0.0", include_video=False)
signsuisse = tfds.load('signsuisse', builder_kwargs=dict(config=config))

# print([d["p.ose"]["data"].shape for d in iter(autsl["train"])])
# print([d["video"].shape for d in iter(autsl["train"])])

# config = SignDatasetConfig(name="include4", version="1.0.0", extra={"PHPSESSID": "hj9co07ct7f5noq529no9u09l4"})
# signtyp = tfds.load(name='sign_typ', builder_kwargs=dict(config=config))
#
# for datum in itertools.islice(signtyp["train"], 0, 10):
#   print(datum['sign_writing'].numpy().decode('utf-8'), datum['video'].numpy().decode('utf-8'))
#
# config = SignDatasetConfig(name="poses_1", version="1.0.0", include_video=False, include_pose="holistic")
# dicta_sign = tfds.load(name='dicta_sign', builder_kwargs={"config": config})

# config = SignDatasetConfig(name="only-annotations5", version="1.0.0", include_video=False, process_video=False, include_pose="holistic")
# dataset = tfds.load(name='sign_bank', builder_kwargs=dict(config=SignDatasetConfig(name="annotations")))
#
# decode_str = lambda s: s.numpy().decode('utf-8')
# for datum in itertools.islice(dataset["train"], 0, 10):
#     hamnosys = decode_str(datum['hamnosys'])
#     glosses = [decode_str(g) for g in datum["glosses"]]
#     print(hamnosys, glosses)


#
# import tensorflow_datasets as tfds
# # noinspection PyUnresolvedReferences
# import sign_language_datasets.datasets
# from sign_language_datasets.datasets.config import SignDatasetConfig
#
# # Populate your access tokens
# TOKENS = {
#     "zenodo_focusnews_token": "TODO",
#     "zenodo_srf_videos_token": "TODO",
#     "zenodo_srf_poses_token": "TODO"
# }
#
# # Load only the annotations, and include path to video files
# config = SignDatasetConfig(name="annotations", version="1.0.0", process_video=False)
# wmtslt = tfds.load(name='wmtslt', builder_kwargs={"config": config, **TOKENS})
#
# # Load the annotations and openpose poses
# config = SignDatasetConfig(name="openpose", version="1.0.0", process_video=False, include_pose='openpose')
# wmtslt = tfds.load(name='wmtslt', builder_kwargs={"config": config, **TOKENS})
#
# # Load the annotations and mediapipe holistic poses
# config = SignDatasetConfig(name="holistic", version="1.0.0", process_video=False, include_pose='holistic')
# wmtslt = tfds.load(name='wmtslt', builder_kwargs={"config": config, **TOKENS})
#
# # Load the full video frames as a tensor
# config = SignDatasetConfig(name="videos", version="1.0.0", process_video=True)
# wmtslt = tfds.load(name='wmtslt', builder_kwargs={"config": config, **TOKENS})
#
decode_str = lambda s: s.numpy().decode('utf-8')
for datum in itertools.islice(signbank["train"], 0, 10):
    print(datum)
    print('\n')






"""WMT-SLT: Data for the WMT shared task on sign language translation"""
import json
import os
from os import path

import requests
import tensorflow_datasets as tfds
import tensorflow as tf
from pose_format import Pose
from pose_format.numpy import NumPyPoseBody
from tensorflow.python.platform.gfile import GFile

from sign_language_datasets.utils.features import PoseFeature
from .utils import (
    get_video_metadata,
    read_mediapipe_surrey_format,
    read_openpose_surrey_format,
    reduce_pose_people,
    convert_srt_time_to_frame,
)
from ..warning import dataset_warning

from ...datasets import SignDatasetConfig

_DESCRIPTION = """
These are Standard German daily news (Tagesschau) and Swiss German weather forecast (Meteo) episodes broadcast and interpreted into Swiss German Sign Language by hearing interpreters (among them, children of Deaf adults, CODA) via Swiss National TV (Schweizerisches Radio und Fernsehen, SRF) (https://www.srf.ch/play/tv/sendung/tagesschau-in-gebaerdensprache?id=c40bed81-b150-0001-2b5a-1e90e100c1c0). For a more extended description of the data, visit https://www.wmt-slt.com/data.
"""

_CITATION = """
@misc{muller_mathias_2022_6637392,
  author       = {Müller Mathias and
                  Ebling Sarah and
                  Camgöz Necati Cihan and
                  Jian Zifan and
                  Battisti Alessia and
                  Tissi Katja and
                  Sidler-Miserez Sandra and
                  Perrollaz Regula and
                  Berger Michèle and
                  Reinhard Sabine and
                  Moryossef Amit and
                  Rios Annette and
                  Bowden Richard and
                  Wong Ryan and
                  Ribback Robin and
                  Schori Severine},
  title        = {{WMT-SLT SRF: Training data for the WMT shared task 
                   on sign language translation (videos, subtitles)}},
  month        = jun,
  year         = 2022,
  note         = {{We additionally acknowledge funding through the 
                   Innosuisse Flagship "Inclusive Information and
                   Communication Technologies" (IICT) (grant
                   agreement no. PFFS-21-47).}},
  publisher    = {Zenodo},
  version      = {V1.2},
  doi          = {10.5281/zenodo.6637392},
  url          = {https://doi.org/10.5281/zenodo.6637392}
}
"""

_POSE_HEADERS = {
    "holistic": path.join(path.dirname(path.realpath(__file__)), "holistic.poseheader"),
    "openpose": path.join(path.dirname(path.realpath(__file__)), "openpose_135.poseheader"),
}

# these IDs change if a new version is released on Zenodo
ZENODO_DEPOSIT_ID_SRF_POSES = 6631275
ZENODO_DEPOSIT_ID_SRF_VIDEOS_SUBTITLES = 6637392
ZENODO_DEPOSIT_ID_FOCUS_NEWS = 6631159


class WMTSLT(tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for wmt slt srf dataset."""

    VERSION = tfds.core.Version("1.2.0")
    RELEASE_NOTES = {
        "1.2.0": "10.5281/zenodo.6637392 Jun 12, 2022",
    }

    BUILDER_CONFIGS = [
        SignDatasetConfig(name="default", process_video=True),
        SignDatasetConfig(name="annotations", process_video=False),
    ]

    def __init__(self, zenodo_srf_poses_token: str, zenodo_srf_videos_token: str, zenodo_focusnews_token: str, **kwargs):
        super().__init__(**kwargs)

        self.zenodo_srf_poses_token = zenodo_srf_poses_token
        self.zenodo_srf_videos_token = zenodo_srf_videos_token
        self.zenodo_focusnews_token = zenodo_focusnews_token

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""

        features = {
            "id": tfds.features.Text(),
            "source": tfds.features.Text(),  # srf|focusnews
            "fps": tf.int32,
            "subtitle": {"text": tfds.features.Text(), "start": tfds.features.Text(), "end": tfds.features.Text(),},
        }

        # Add poses if requested
        if self._builder_config.include_pose is not None:
            if self._builder_config.include_pose == "openpose":
                pose_shape = (None, 1, 135, 2)
            elif self._builder_config.include_pose == "holistic":
                pose_shape = (None, 1, 203, 3)
            else:
                raise Exception("Unknown pose format")

            features["pose"] = PoseFeature(shape=pose_shape)

        if self._builder_config.process_video:
            features["video"] = self._builder_config.video_feature((720, 640))
            raise Exception("Video processing is not currently supported")
        else:
            features["video"] = tfds.features.Text()

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features),
            homepage="https://www.wmt-slt.com/data",
            supervised_keys=None,
            citation=_CITATION,
        )

    def download_zenodo_deposit(self, dl_manager: tfds.download.DownloadManager, zenodo_deposit_id: int, zenodo_token: str):
        # using private link to set cookies
        record_link = f"https://zenodo.org/record/{zenodo_deposit_id}?token={zenodo_token}"
        record_request = requests.get(record_link)
        cookie = record_request.headers["Set-Cookie"].split(";")[0]

        # query the API for information about the deposit, extract the data link from the response
        record_api_link = f"https://zenodo.org/api/records/{zenodo_deposit_id}"
        headers = {"Cookie": cookie}
        record_api_response = json.loads(requests.get(record_api_link, headers=headers).text)

        zip_file = record_api_response["files"][0]["links"]["self"]
        archive = dl_manager.download_and_extract(tfds.download.Resource(url=zip_file, headers=headers))
        return archive

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        dataset_warning(self)

        srf_poses = self.download_zenodo_deposit(dl_manager, ZENODO_DEPOSIT_ID_SRF_POSES, self.zenodo_srf_poses_token)
        srf_poses_dir = path.join(srf_poses, "srf", "parallel")

        srf_videos = self.download_zenodo_deposit(dl_manager, ZENODO_DEPOSIT_ID_SRF_VIDEOS_SUBTITLES, self.zenodo_srf_videos_token)
        srf_parallel_dir = path.join(srf_videos, "srf", "parallel")

        focusnews = self.download_zenodo_deposit(dl_manager, ZENODO_DEPOSIT_ID_FOCUS_NEWS, self.zenodo_focusnews_token)
        focusnews_dir = path.join(focusnews, "focusnews")

        datasets = {
            "srf": {
                "videos": path.join(srf_parallel_dir, "videos"),  # "srf.2020-03-13.mp4"
                "subtitles": path.join(srf_parallel_dir, "subtitles"),  # "srf.2020-03-13.srt"
                "openpose": path.join(srf_poses_dir, "openpose"),  # "srf.2020-03-13.openpose.tar.xz"
                "mediapipe": path.join(srf_poses_dir, "mediapipe"),  # "srf.2020-03-13.mediapipe.tar.xz"
            },
            "focusnews": {
                "videos": path.join(focusnews_dir, "videos"),  # "focusnews.043.mp4"
                "subtitles": path.join(focusnews_dir, "subtitles"),  # "focusnews.043.srt"
                "openpose": path.join(focusnews_dir, "openpose"),  # "focusnews.043.openpose.tar.xz"
                "mediapipe": path.join(focusnews_dir, "mediapipe"),  # "focusnews.043.mediapipe.tar.xz"
            },
        }

        print(f"SRF data stored in {srf_parallel_dir} and {srf_poses_dir}")
        print(f"FocusNews data stored in {focusnews}")

        return {"train": self._generate_examples(datasets)}

    def _generate_examples(self, datasets):
        """Yields examples."""

        try:
            import srt

        except ImportError:
            raise ImportError("Please install srt with: pip install srt")

        for dataset_id, directories in datasets.items():
            names = [n[: -len(".mp4")] for n in os.listdir(directories["videos"])]
            for name in names:
                video_path = path.join(directories["videos"], f"{name}.mp4")
                subtitles_path = path.join(directories["subtitles"], f"{name}.srt")

                video_metadata = get_video_metadata(video_path)
                fps = video_metadata["fps"]
                width = video_metadata["width"]
                height = video_metadata["height"]

                # # Load video
                # video = None
                # if self.builder_config.process_video:
                #     video =

                # Load poses
                pose = None
                if self.builder_config.include_pose is not None:
                    if self.builder_config.include_pose == "openpose":
                        openpose_path = path.join(directories["openpose"], f"{name}.openpose.tar.xz")
                        pose = read_openpose_surrey_format(openpose_path, fps=fps, width=width, height=height)

                    if self.builder_config.include_pose == "holistic":
                        mediapipe_path = path.join(directories["mediapipe"], f"{name}.mediapipe.tar.xz")
                        pose = read_mediapipe_surrey_format(mediapipe_path, fps=fps, width=width, height=height)

                    reduce_pose_people(pose)

                with GFile(subtitles_path, "r") as f:
                    subtitles = srt.parse(f.read())

                for i, s in enumerate(subtitles):
                    datum = {
                        "id": f"{name}_{i}",
                        "source": dataset_id,
                        "fps": fps,
                        "subtitle": {"start": str(s.start), "end": str(s.end), "text": s.content},
                    }

                    frame_start = convert_srt_time_to_frame(s.start, fps)
                    frame_end = convert_srt_time_to_frame(s.end, fps)

                    if self.builder_config.process_video:
                        pass  # TODO get the relevant video frames
                    else:
                        datum["video"] = video_path

                    if pose is not None:
                        new_pose_data = pose.body.data[frame_start:frame_end]
                        new_pose_confidence = pose.body.confidence[frame_start:frame_end]
                        new_pose_body = NumPyPoseBody(data=new_pose_data, confidence=new_pose_confidence, fps=fps)
                        datum["pose"] = Pose(header=pose.header, body=new_pose_body)

                    yield datum["id"], datum



"""Helper file to crawl TheLanguageArchive and create an up-to-date index of the ngt dataset"""

import json
import re
import os
import sys
import urllib.request

from tqdm import tqdm
from typing import Tuple, Optional


"""
Notes about files and filenames:

The corpus consists of the following files:
- ELAN (*.eaf)
- videos (*.mpg)
- audio (*.wav)

Not all of these file types are present for all corpus items.

All filenames start with an example ID, examples:
- CNGT0100_r3.eaf
- CNGT0100_S002_b.mpg
- CNGT0100.wav

## ELAN Files

The filenames of ELAN files includes the release version, for instance "r3" for the third release. The ELAN XML
also contains references to all video files that are linked to these particular annotations. Example:

<ANNOTATION_DOCUMENT>
    <HEADER MEDIA_FILE="" TIME_UNITS="milliseconds">
        <MEDIA_DESCRIPTOR MEDIA_URL="file:///F:/CNGT MPEG-1 body/CNGT0100_S002_b.mpg" MIME_TYPE="video/mpeg" RELATIVE_MEDIA_URL="../../../../CNGT MPEG-1 body/CNGT0092_S002_b.mpg"/>
        <MEDIA_DESCRIPTOR MEDIA_URL="file:///F:/CNGT MPEG-1 body/CNGT0100_S001_b.mpg" MIME_TYPE="video/mpeg" RELATIVE_MEDIA_URL="../../../../CNGT MPEG-1 body/CNGT0092_S001_b.mpg"/>

## Videos

Filenames of videos have the following structure:

[example ID]_[speaker ID]_[view identifier].mpg

Example: CNGT0100_S002_b.mpg

The speaker ID is a unique identifier for a specific person, unlike in the public DGS, where identifiers "a" and "b"
refer to different people depending on the corpus example.

There are 4 types of camera views:
- "f": face of speaker, frontal view, resolution: 352 x 288
- "b": body of speaker, including the face, frontal view, resolution: 352 x 288
- "t": top, the speaker filmed from above, resolution: 352 x 288
- no view identifier and no speaker (i.e. "CNGT0100.mpg" for example): wider-angle video that
  shows the entire scene, resolution: 704 x 288

As far as I can see, "f" and "b" are present for all corpus examples, and for all speakers. "t" and the entire scene
are missing sometimes.

Framerate of videos: 25 fps

## Restricted files

Some files are restricted and cannot be found by the crawler below. This does not mean that all files are restricted,
in some cases the ELAN file is public (albeit empty) but the videos are restricted.

Example: https://archive.mpi.nl/tla/islandora/object/tla%3A1839_00_0000_0000_0009_2D68_9

ELAN files for which the videos are not public are marked with "NP" in the file name, example: "CNGT0020_r3_NP.eaf".

The opposite case is also possible: sometimes only a video of the combined or "t" view is available, while everything else
including the ELAN file is restricted.

Example: CNGT0314 here: https://archive.mpi.nl/tla/islandora/object/tla%3A1839_00_0000_0000_0009_0016_1

List of all such exceptions currently: {'CNGT0025', 'CNGT0314', 'CNGT1542', 'CNGT0020', 'CNGT0876'}
"""


base_path = "https://archive.mpi.nl"


def get_sub_pages(html):
    return [x for x in re.findall('<strong class="field-content"><a href="(.*)">.*</a></strong>', html)]


def create_flat_download_dict() -> dict:

    # Iterate over pages
    next_page = "/tla/islandora/object/tla%3A1839_00_0000_0000_0004_F3D5_A"
    sub_pages = []
    while next_page is not None:
        with urllib.request.urlopen(base_path + next_page) as response:
            html = response.read().decode("utf-8")

            next_search = re.search('<a href="(.*?)">next</a>', html)
            next_page = next_search.group(1) if next_search else None

            sub_pages += get_sub_pages(html)

    # Iterate over sub pages
    download_pages = []
    for page in tqdm(sub_pages):
        print(base_path + page)
        with urllib.request.urlopen(base_path + page) as response:
            html = response.read().decode("utf-8")
            download_pages += get_sub_pages(html)

    # Iterate over download pages (creates a flat list of files)

    flat_download_dict = {}

    for page in tqdm(download_pages):
        print(base_path + page)
        with urllib.request.urlopen(base_path + page) as response:
            html = response.read().decode("utf-8")

            files = re.findall(
                'title="view item">(.*)</a></div>\n<div class="flat-compound-buttons">\n<div class="flat-compound-download"><a href="(.*)" class="flat-compound-download" title="download file"></a>',
                html,
            )
            for file in files:
                flat_download_dict[file[0]] = base_path + file[1]

    return flat_download_dict


def get_id_from_filename(filename: str) -> str:
    """

    :param filename:
    :return:
    """
    if "_" not in filename:
        # assume combined view
        example_id = filename.split(".")[0]
    else:
        example_id = filename.split("_")[0]
    assert example_id.startswith("CNGT")
    return example_id


def get_info_from_mpg_filename(filename: str) -> Tuple[Optional[str], str]:
    """

    :param filename:
    :return:
    """
    if "S" not in filename:
        # assume combined view
        speaker_id = None
    else:
        speaker_id = filename.split("_")[1].split(".")[0]
        assert speaker_id.startswith("S"), "Speaker ID of filename '%s' does not start with S: '%s'" % (filename, speaker_id)

    if "_b" in filename:
        view = "b"
    elif "_f" in filename:
        view = "f"
    elif "_t" in filename:
        view = "t"
    else:
        # assume combined view
        view = "c"

    return speaker_id, view


def create_structured_download_dict(
    force_rebuild: bool = False, flat_json_path: str = "ngt_flat.json", structured_json_path: str = "ngt.json"
):
    """

    :param force_rebuild:
    :param flat_json_path:
    :param structured_json_path:
    :return:
    """
    # open or create flat download dict

    if os.path.exists(flat_json_path) and not force_rebuild:
        print("Opening existing flat download dict: '%s'" % flat_json_path)
        with open(flat_json_path, "r") as infile:
            flat_download_dict = json.load(infile)
    else:
        print("Rebuilding flat download dict: '%s'" % flat_json_path)
        flat_download_dict = create_flat_download_dict()

        with open(flat_json_path, "w") as outfile:
            json.dump(flat_download_dict, outfile)

    # group files by corpus example ID

    structured_download_dict = {}

    for filename, url in flat_download_dict.items():

        if "-" in filename:
            print("Found hyphen in filename: '%s'. Will replace with underscore." % filename)
            filename = filename.replace("-", "_")

        example_id = get_id_from_filename(filename)

        if example_id not in structured_download_dict.keys():
            structured_download_dict[example_id] = {}

        if ".wav" in filename:
            structured_download_dict[example_id]["audio"] = url
        elif ".eaf" in filename:
            structured_download_dict[example_id]["eaf"] = url
        elif ".mpg" in filename:
            speaker_id, view = get_info_from_mpg_filename(filename)

            if view == "c":
                video_id = "video_c"
            else:
                video_id = "video_" + speaker_id + "_" + view

            structured_download_dict[example_id][video_id] = url
        else:
            print("Unexpected filename: %s" % filename)
            sys.exit()

    with open(structured_json_path, "w") as outfile:
        print("Writing structured download dict '%s'." % "ngt.json")
        json.dump(structured_download_dict, outfile)


if __name__ == "__main__":
    create_structured_download_dict(flat_json_path="ngt_flat.json", structured_json_path="ngt.json")



"""ocr tests"""
import os
from unittest import TestCase
import cv2

from sign_language_datasets.utils.signwriting.ocr import image_to_fsw


class TestOCR(TestCase):
    def test_should_extract_fsw_from_image(self):
        dirname = os.path.dirname(__file__)
        img_path = os.path.join(dirname, "assets/sign.png")

        img_rgb = cv2.imread(img_path)
        symbols = ["S1f520", "S1f528", "S23c04", "S23c1c", "S2fb04", "S2ff00", "S33b10"]

        self.assertEqual(
            image_to_fsw(img_rgb, symbols),
            "M239x127S2ff00043x057S23c04071x118S23c1c028x118S1f520062x100S1f528035x100S2fb04054x181S33b10054x083",
        )



"""A Swiss Sign Language Lexicon, combines all three Swiss sign languages: the German-Swiss Sign Language (DSGS), the Langue des Signes Française (LSF) and the Lingua Italiana dei Segni (LIS). ."""
import json
import re
import string

import tensorflow_datasets as tfds
from tqdm import tqdm

from ..warning import dataset_warning
from ...datasets import SignDatasetConfig

_DESCRIPTION = """
Ein umfangreiches interaktives Lexikon der drei Gebärdensprachen der Schweiz (Deutschschweizerische Gebärdensprache DSGS, Langue des Signes Française LSF und Lingua Italiana dei Segni LIS). Herausgegeben vom Schweizerischen Gehörlosenbund (SGB-FSS)."""

# TODO(signsuisse): BibTeX citation
_CITATION = """
"""

SITE_URL = "https://signsuisse.sgb-fss.ch"


class SignSuisse(tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for signsuisse dataset."""

    MAX_SIMULTANEOUS_DOWNLOADS = 100  # Default is 50, but it is slow
    VERSION = tfds.core.Version("1.0.0")
    RELEASE_NOTES = {
        "1.0.0": "Initial crawl.",
    }

    BUILDER_CONFIGS = [
        SignDatasetConfig(name="default", include_video=True),
        SignDatasetConfig(name="annotations", include_video=False),
    ]

    def _info(self) -> tfds.core.DatasetInfo:
        """Return s the dataset metadata."""

        features = {
            "id": tfds.features.Text(),
            "name": tfds.features.Text(),
            "category": tfds.features.Text(),
            "spokenLanguage": tfds.features.Text(),
            "signedLanguage": tfds.features.Text(),
            "url": tfds.features.Text(),
            "paraphrase": tfds.features.Text(),
            "definition": tfds.features.Text(),
            "exampleText": tfds.features.Text(),
            "exampleVideo": tfds.features.Text(),
        }

        if self._builder_config.include_video and self._builder_config.process_video:
            features["video"] = self._builder_config.video_feature((640, 480))
        else:
            features["video"] = tfds.features.Text()

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features),
            homepage=SITE_URL,
            supervised_keys=None,
            citation=_CITATION,
        )

    def _list_all_lexicon_items(self, dl_manager: tfds.download.DownloadManager):
        language = ['de', 'fr', 'it']
        dl_links = [f"https://signsuisse.sgb-fss.ch/index.php?eID=sitemap&lang={l}&format=json" for l in language]

        indexes = dl_manager.download(dl_links)
        results = []
        for index in indexes:
            with open(index, "r", encoding="utf-8") as f:
                index = json.load(f)
            results.extend(index)

        return results

    def _parse_item(self, item, item_page):
        item["name"] = item["name"].replace("   ", " ").replace("  ", " ")
        with open(item_page, "r", encoding="utf-8") as f:
            html = f.read()

        # Verify that the page matches the item
        title = re.search(r"<h1.*?>(.*?)</h1>", html).group(1).strip()
        title = title.replace("&amp;", "&")
        if title != item["name"]:
            raise ValueError("Title does not match item name")

        assert title == item["name"]

        example = re.search(r"Beispiel</h2> <p>(.*?)</p>.*?<video.*?src=\"(.*?)\"", html)
        if example is not None:
            example_text = example.group(1).strip()
            example_video = SITE_URL + example.group(2).strip()
        else:
            # This happens on: https://signsuisse.sgb-fss.ch/de/lexikon/g/shakespeare-william/
            # Because the page is formatted differently. I requested a change to the page.
            example_text = ""
            example_video = ""

        if item["uid"] in ["131220", "131073", "120731", "115584", "121152", "121125", "126042"]:
            # This is a special case where the video is not available.
            return None

        video = SITE_URL + re.search(r"<video id=\"video-main\".*?src=\"(.*?)\"", html).group(1).strip()
        paraphrase_match = re.search(r"Umschreibung</h2> <p>(.*?)</p>", html)
        paraphrase = paraphrase_match.group(1).strip() if paraphrase_match else ""
        definition_match = re.search(r"Definition</h2> <p>(.*?)</p>", html)
        definition = definition_match.group(1).strip() if definition_match else ""
        spoken_language = item["sprache"]
        url_path = re.search(rf"<a href=\"(\/{spoken_language}\/.*?)\"", html).group(1).strip()

        return {
            "id": item["uid"],
            "name": item["name"],
            "category": item["kategorie"],
            "spokenLanguage": spoken_language,
            "signedLanguage": "ch-" + spoken_language,
            "url": SITE_URL + url_path,
            "paraphrase": paraphrase,
            "definition": definition,
            "exampleText": example_text,
            "exampleVideo": example_video,
            "video": video
        }

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        dataset_warning(self)
        print(
            "The lexicon is available free of charge, so we look forward to your donation! https://www.sgb-fss.ch/spenden/jetzt-spenden/")

        lexicon_items = self._list_all_lexicon_items(dl_manager)
        print("Found", len(lexicon_items), "lexicon items.")
        item_urls = [item["link"] for item in lexicon_items]
        # Item URLS are actually too long. We need to shorten them.
        item_urls = [url[:url.find("&tx_issignsuisselexikon_anzeige%5Baction")] for url in item_urls]
        items_pages = dl_manager.download(item_urls)

        data = []
        for item, item_page in zip(lexicon_items, items_pages):
            try:
                item = self._parse_item(item, item_page)
                if item is not None:
                    data.append(item)
            except Exception as e:
                print("Failed to parse item")
                print(item)
                print(item_page)
                print(e)
                print("\n\n\n\n\n\n")
                # raise e

        # Download videos if requested
        if self._builder_config.include_video:
            video_urls = [item["video"] for item in data]
            videos = dl_manager.download(video_urls)
            for datum, video in zip(data, videos):
                datum["video"] = video
                if not self._builder_config.process_video:
                    datum["video"] = str(datum["video"])

        return {"train": self._generate_examples(data)}

    def _generate_examples(self, data):
        for datum in data:
            yield datum["id"], datum



import tensorflow_datasets as tfds
import sign_language_datasets.datasets
from sign_language_datasets.datasets.config import SignDatasetConfig

import itertools
import numpy as np

import torch


class TFDSTorchDataset(torch.utils.data.IterableDataset):
    def __init__(self, tf_dataset):
        self.dataset = tf_dataset

    def __len__(self):
        return len(self.dataset)

    def torch_obj(self, obj):
        if isinstance(obj, dict):
            return {k: self.torch_obj(v) for k, v in obj.items()}
        if isinstance(obj, np.ndarray):
            return torch.from_numpy(obj)
        return obj

    def __iter__(self):
        for datum in self.dataset:
            yield self.torch_obj(tfds.as_numpy(datum))


if __name__ == "__main__":
    # Fast download and load dataset using TFDS
    config = SignDatasetConfig(name="holistic-poses", version="1.0.0", include_video=False, include_pose="holistic")
    dicta_sign = tfds.load(name="dicta_sign", builder_kwargs={"config": config})

    # Convert to torch dataset
    train_dataset = TFDSTorchDataset(dicta_sign["train"])

    for datum in itertools.islice(train_dataset, 0, 10):
        print(datum)



"""bsl_corpus dataset."""

import os

import tensorflow_datasets as tfds
from tensorflow_datasets.core import download
from tensorflow_datasets.core import utils

from . import bsl_corpus_utils
from ..warning import dataset_warning
from ...datasets.config import SignDatasetConfig

_DESCRIPTION = """
A corpus of British Sign Language.
"""

_CITATION = """\
@inproceedings{cormier:12033:sign-lang:lrec,
  author    = {Cormier, Kearsy and Fenlon, Jordan and Johnston, Trevor and Rentelis, Ramas and Schembri, Adam and Rowley, Katherine and Adam, Robert and Woll, Bencie},
  title     = {From Corpus to Lexical Database to Online Dictionary: Issues in annotation of the {BSL} Corpus and the Development of {BSL} {SignBank}},
  pages     = {7--12},
  editor    = {Crasborn, Onno and Efthimiou, Eleni and Fotinea, Stavroula-Evita and Hanke, Thomas and Kristoffersen, Jette and Mesch, Johanna},
  booktitle = {Proceedings of the {LREC2012} 5th Workshop on the Representation and Processing of Sign Languages: Interactions between Corpus and Lexicon},
  maintitle = {8th International Conference on Language Resources and Evaluation ({LREC} 2012)},
  publisher = {{European Language Resources Association (ELRA)}},
  address   = {Istanbul, Turkey},
  day       = {27},
  month     = may,
  year      = {2012},
  language  = {english},
  url       = {https://www.sign-lang.uni-hamburg.de/lrec/pub/12033.pdf}
}
"""

_HOMEPAGE = "https://bslcorpusproject.org/"

_VIDEO_RESOLUTION = (640, 360)

_FRAMERATE = 25


class BslCorpus(tfds.core.GeneratorBasedBuilder):
    """
  DatasetBuilder for BSL corpus dataset.

  The dataset currently loads annotations only, and ignores video files.
  """

    VERSION = tfds.core.Version("1.0.0")
    RELEASE_NOTES = {
        # Third release of annotations:
        # https://bslcorpusproject.org/wp-content/uploads/Notes-to-the-3rd-release-of-BSL-Corpus-annotations.pdf
        "3.0.0": "Third release.",
    }

    BUILDER_CONFIGS = [
        SignDatasetConfig(name="default", include_video=False, include_pose=None),
        SignDatasetConfig(name="annotations", include_video=False, include_pose=None),
    ]

    def __init__(self, bslcp_username: str, bslcp_password: str, **kwargs):

        super(BslCorpus, self).__init__(**kwargs)

        self.bslcp_username = bslcp_username
        self.bslcp_password = bslcp_password

    def _make_download_manager(self, download_dir, download_config):
        """Creates a new download manager object."""
        download_dir = download_dir or os.path.join(self._data_dir_root, "downloads")
        extract_dir = download_config.extract_dir or os.path.join(download_dir, "extracted")
        manual_dir = download_config.manual_dir or os.path.join(download_dir, "manual")

        if download_config.register_checksums:
            # Note: Error will be raised here if user try to record checksums
            # from a `zipapp`
            # noinspection PyTypeChecker
            register_checksums_path = utils.to_write_path(self._checksums_path)
        else:
            register_checksums_path = None

        return bsl_corpus_utils.DownloadManagerBslCorpus(
            download_dir=download_dir,
            extract_dir=extract_dir,
            manual_dir=manual_dir,
            url_infos=self.url_infos,
            manual_dir_instructions=self.MANUAL_DOWNLOAD_INSTRUCTIONS,
            force_download=(download_config.download_mode == download.GenerateMode.FORCE_REDOWNLOAD),
            force_extraction=(download_config.download_mode == download.GenerateMode.FORCE_REDOWNLOAD),
            force_checksums_validation=download_config.force_checksums_validation,
            register_checksums=download_config.register_checksums,
            register_checksums_path=register_checksums_path,
            verify_ssl=download_config.verify_ssl,
            dataset_name=self.name,
            username=self.bslcp_username,
            password=self.bslcp_password,
            max_retries=5,
        )

    def _info(self) -> tfds.core.DatasetInfo:
        """
        Returns the dataset metadata.

        Note: The corpus has much more metadata about each record. This meta data is currently not exposed
              in this data loader. See `bsl_corpus_utils._stream_file_from_container_url` which also
              returns a metadata dictionary.
        """
        features = {
            "id": tfds.features.Text(),
            "paths": {"eaf": tfds.features.Sequence(tfds.features.Text(), length=None),},
        }

        if self._builder_config.include_video:
            raise NotImplementedError("Videos are currently not available for the BSL corpus.")

        if self._builder_config.include_pose is not None:
            raise NotImplementedError("Poses are currently not available for the BSL corpus.")

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features),
            homepage=_HOMEPAGE,
            supervised_keys=None,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        """Returns SplitGenerators."""
        dataset_warning(self)

        records_iterator = bsl_corpus_utils.generate_download_links(
            username=self.bslcp_username,
            password=self.bslcp_password,
            number_of_records=None,  # no limit on num records
            renew_user_token_every_n_pages=5,
        )

        processed_data = {}

        for records_per_result_page in records_iterator:
            index_data = {}

            for record in records_per_result_page:
                _id = record["record_id"]
                eaf_urls = record["downloads"].get("EAF file", None)

                if eaf_urls is None or len(eaf_urls) == 0:
                    continue

                for eaf_index, eaf_url in enumerate(eaf_urls):

                    datum = {"eaf_" + str(eaf_index): eaf_url}
                    index_data[_id] = datum

            urls = {url: url for datum in index_data.values() for url in datum.values() if url is not None}

            local_paths = dl_manager.download(urls)

            processed_data_per_results_page = {
                _id: {k: local_paths[v] if v is not None else None for k, v in datum.items()} for _id, datum in index_data.items()
            }

            processed_data.update(processed_data_per_results_page)

        return [tfds.core.SplitGenerator(name=tfds.Split.TRAIN, gen_kwargs={"data": processed_data})]

    def _generate_examples(self, data):
        """ Yields examples. """

        for _id, datum in list(data.items()):

            eaf_files_as_list = []
            eaf_keys = [k for k in datum.keys() if k.startswith("eaf")]

            for eaf_key in sorted(eaf_keys):
                eaf_files_as_list.append(str(datum[eaf_key]))

            features = {
                "id": _id,
                "paths": {"eaf": eaf_files_as_list},
            }

            if self._builder_config.include_video:
                raise NotImplementedError("Videos are currently not available for the BSL corpus.")

            if self._builder_config.include_pose == "openpose":
                raise NotImplementedError("OpenPose poses are currently not available for the BSL corpus.")

            if self._builder_config.include_pose == "holistic":
                raise NotImplementedError("Holistic poses are currently not available for the BSL corpus.")

            yield _id, features



"""A Brazilian Sign Language (Libras) lexicon."""
import json

import tensorflow as tf
import tensorflow_datasets as tfds

from ..warning import dataset_warning
from ...datasets import SignDatasetConfig

_DESCRIPTION = """
A lexicon for Brazilian Sign Language (Libras), including videos, and SignWriting transcription.
"""

# TODO(swojs_glossario): BibTeX citation
_CITATION = """
"""


def get_media_id(datum):
    if "o:media" not in datum or len(datum["o:media"]) == 0:
        return None
    return datum["o:media"][0]["o:id"]


def get_transcription(datum):
    vals = []
    if "oc:transcription" in datum:
        vals = datum["oc:transcription"]
    elif "http://purl.org/linguistics/gold:writtenRealization" in datum:
        vals = datum["http://purl.org/linguistics/gold:writtenRealization"]

    return [t["@value"] for t in vals]


class SwojsGlossario(tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for swojs_glossario dataset."""

    VERSION = tfds.core.Version("1.0.0")
    RELEASE_NOTES = {
        "1.0.0": "Initial release.",
    }

    BUILDER_CONFIGS = [
        SignDatasetConfig(name="default", include_video=True),
        SignDatasetConfig(name="annotations", include_video=False),
    ]

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""

        features = {
            "id": tf.int32,
            "title": tfds.features.Text(),
            "created_date": tfds.features.Text(),
            "modified_date": tfds.features.Text(),
            "sign_writing": tfds.features.Sequence(tfds.features.Text()),
            "spoken_language": tfds.features.Sequence(tfds.features.Text()),
            "signed_language": tfds.features.Sequence(tfds.features.Text()),
        }

        if self._builder_config.include_video and self._builder_config.process_video:
            features["video"] = self._builder_config.video_feature((1920, 1080))
        else:
            features["video"] = tfds.features.Text()

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features),
            homepage="http://swojs.ibict.br/portal/s/swojs/page/glossario",
            supervised_keys=None,
            citation=_CITATION,
        )

    def get_media_dict(self, media_path):
        with open(media_path, "r", encoding="utf-8") as g:
            media = json.load(g)
            return {m["o:id"]: m["o:original_url"] for m in media}

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        dataset_warning(self)

        annotations_path, media_path = dl_manager.download(
            ["http://swojs.ibict.br/portal/api/items?page", "http://swojs.ibict.br/portal/api/media?page"]
        )

        media_dict = self.get_media_dict(media_path)

        local_videos = {}
        if self._builder_config.include_video and self._builder_config.process_video:
            with open(annotations_path, "r", encoding="utf-8") as f:
                annotations = json.load(f)

                medias = [_id for _id in [get_media_id(a) for a in annotations] if _id is not None]
                media_urls = [media_dict[_id] for _id in medias]

                video_paths = dl_manager.download(media_urls)
                local_videos = {k: v for k, v in zip(medias, video_paths)}

        return {"train": self._generate_examples(annotations_path, local_videos, media_dict)}

    def _generate_examples(self, annotations_path, local_videos, media_dict):
        """Yields examples."""

        with open(annotations_path, "r", encoding="utf-8") as f:
            data = json.load(f)
            for datum in data:
                features = {
                    "id": datum["o:id"],
                    "title": datum["o:title"] if datum["o:title"] is not None else "",
                    "created_date": datum["o:created"]["@value"],
                    "modified_date": datum["o:modified"]["@value"],
                    "sign_writing": get_transcription(datum),
                    "spoken_language": [],
                    "signed_language": [],
                }

                if "http://www.lingvoj.org/ontology:originalLanguage" in datum:
                    features["spoken_language"] = [t["@value"] for t in datum["http://www.lingvoj.org/ontology:originalLanguage"]]

                if "http://www.lingvoj.org/ontology:targetLanguage" in datum:
                    features["signed_language"] = [t["@value"] for t in datum["http://www.lingvoj.org/ontology:targetLanguage"]]

                media_id = get_media_id(datum)
                if media_id in local_videos:
                    features["video"] = local_videos[media_id]
                elif media_id is not None:
                    features["video"] = media_dict[media_id]
                else:
                    features["video"] = ""

                yield str(features["id"]), features



"""signsuisse dataset."""

import tensorflow_datasets as tfds
from . import signsuisse


class SignSuisseTest(tfds.testing.DatasetBuilderTestCase):
    """Tests for signsuisse dataset."""

    # TODO(signsuisse):
    DATASET_CLASS = signsuisse.SignSuisse
    SPLITS = {
        "train": 3,  # Number of fake train example
        "test": 1,  # Number of fake test example
    }

    # If you are calling `download/download_and_extract` with a dict, like:
    #   dl_manager.download({'some_key': 'http://a.org/out.txt', ...})
    # then the tests needs to provide the fake output paths relative to the
    # fake data directory
    # DL_EXTRACT_RESULT = {'some_key': 'output_file1.txt', ...}


if __name__ == "__main__":
    tfds.testing.test_main()



"""ASLG-PC12: Synthetic English-ASL Gloss Parallel Corpus 2012"""

import tensorflow_datasets as tfds
from tensorflow.io.gfile import GFile

from sign_language_datasets.datasets.warning import dataset_warning

_DESCRIPTION = """
A large synthetic collection of parallel English and ASL-Gloss texts.
There are two string features: text, and gloss.
"""

_CITATION = """
@inproceedings{othman2012english,
  title={English-asl gloss parallel corpus 2012: Aslg-pc12},
  author={Othman, Achraf and Jemni, Mohamed},
  booktitle={5th Workshop on the Representation and Processing of Sign Languages: Interactions between Corpus and Lexicon LREC},
  year={2012}
}
"""

_GLOSS_URL = "https://www.achrafothman.net/aslsmt/corpus/sample-corpus-asl-en.asl"
_TEXT_URL = "https://www.achrafothman.net/aslsmt/corpus/sample-corpus-asl-en.en"


class AslgPc12(tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for aslg_pc12 dataset."""

    VERSION = tfds.core.Version("0.0.1")
    RELEASE_NOTES = {"0.0.1": "Sample of the full corpus"}

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict({"gloss": tfds.features.Text(), "text": tfds.features.Text()}),
            supervised_keys=("gloss", "text"),  # Set to `None` to disable
            homepage="https://achrafothman.net/site/asl-smt/",
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        """Returns SplitGenerators."""
        dataset_warning(self)

        gloss_path, text_path = dl_manager.download([_GLOSS_URL, _TEXT_URL])

        return {
            "train": self._generate_examples(gloss_path, text_path),
        }

    def _generate_examples(self, gloss_path: str, text_path: str):
        """Yields examples."""

        with GFile(gloss_path, "r") as gloss_f:
            with GFile(text_path, "r") as text_f:
                for i, (gloss, text) in enumerate(zip(gloss_f, text_f)):
                    yield i, {"gloss": gloss, "text": text}



import re
import io
import contextlib
import requests
import os
import pympi

import tensorflow_datasets as tfds

import tensorflow as tf
from tensorflow_datasets.core.download import downloader
from tensorflow_datasets.core import utils
from tensorflow_datasets.core import units
from tensorflow_datasets.core.download import checksums as checksums_lib

from requests.adapters import HTTPAdapter, Retry
from typing import Optional, List, Dict, Any, Tuple, Iterator, Iterable


def login_with_credentials(username: str, password: str, base_url: str = "http://digital-collections.ucl.ac.uk/") -> str:
    """

    :param username:
    :param password:
    :param base_url:
    :return:
    """
    # Get initial page to generate a user token

    initial_url = base_url + "R&?local_base=BSLCP"
    initial_response = requests.get(initial_url)

    downloader._assert_status(initial_response)
    initial_response = initial_response.text

    # Perform the login operation for the retrieved user token

    redirect_url = re.search(r"<a href=\"(.*?)\">Researcher Login</a>", initial_response).groups()[0]

    post_data = {
        "func": "login",
        "calling_systemv": "DIGITOOL",
        "selfreg": "",
        "bor_id": username,
        "bor_verification": password,
        "institute": "",
        "url": redirect_url,
    }

    login_response = requests.post(base_url + "pds", data=post_data).text

    login_redirect_response = login_response
    for i in range(2):
        login_redirect_uri = re.search(r"<body onload = \"location = '/(.*?)'", login_redirect_response).groups()[0]
        login_redirect_response = requests.get(base_url + login_redirect_uri, allow_redirects=True).text

    user_token = re.search(r"URL=.*?/R/(.*?)-", login_redirect_response).groups()[0]

    return user_token


def get_search_response(user_token: str, base_url: str = "http://digital-collections.ucl.ac.uk/") -> Tuple[str, str]:
    """

    :param user_token:
    :param base_url:
    :return:
    """
    # The way this works is we have a TOKEN and a request identifier.
    # Seems like they store both, and show the relevant info for the identifier.

    search_url = base_url + "R/" + user_token + "-00001"
    post_data = {
        "func": "search-advanced-go",
        "file_format_code": "WEX",
        "local_base": "BSLCP",
        "mode": "1",
        "find_code1": "WRD",
        "request1": "bslcp",
        "find_operator": "AND",
        "find_code2": "WRD",
        "request2": "",
        "find_operator2": "AND",
        "find_code3": "WRD",
        "request3": "",
        "media_type": "ALL",
        "selected_otype": "",
        "selected_tag": "0",
    }

    search_response_text = requests.post(search_url, data=post_data).text

    return search_url, search_response_text


def generate_download_links(
    username: str,
    password: str,
    number_of_records: Optional[int] = None,
    base_url: str = "http://digital-collections.ucl.ac.uk/",
    renew_user_token_every_n_pages: int = 5,
) -> Iterator[List[Dict]]:
    """
    Yields 20 search results at a time

    :param username:
    :param password:
    :param number_of_records:
    :param base_url:
    :param renew_user_token_every_n_pages:
    :return:
    """
    if number_of_records is None:
        # do an initial search to find the total number of records
        user_token = login_with_credentials(username=username, password=password, base_url=base_url)

        # We are logged in, now let's search
        _, search_response_text = get_search_response(user_token=user_token, base_url=base_url)

        number_of_records = int(re.search(r"Records.*?of\s*(\d*)", search_response_text).groups()[0])

    for i in range(1, number_of_records, 20):  # Page size is 20, no control over that that I have seen

        if i == 1 or i % renew_user_token_every_n_pages == 0:
            # generate a new user token for this result page
            user_token = login_with_credentials(username=username, password=password, base_url=base_url)

            # search again with new user token
            search_url, _ = get_search_response(user_token=user_token, base_url=base_url)

        results_page_response = ""  # all pages are concatenated into a long string

        results_page_url = f"{search_url}?func=results-jump-page&set_entry={i}&result_format=001"
        results_page_response += requests.get(results_page_url).text

        index_data = []

        last_index = 0
        while True:
            try:
                start_section = results_page_response.index("<!-- START_SECTION --><!-- body-line -->", last_index)
            except ValueError:  # substring not found
                break
            end_section = results_page_response.index("<!-- END_SECTION --><!-- body-line -->", last_index)
            last_index = end_section + 1

            cells = re.findall(r"<TD.*?>(.*?)</TD>", results_page_response[start_section:end_section])
            downloads = re.findall(r"<A.*?open_window_delivery\(\"(.*?)\".*?ALT=\"(.*?)\"", cells[3])

            download_dict = {"Quicktime movie": [], "EAF file": [], "MP4 file": [], "MPEG file": [], "Unknown": []}

            for url, file_type in downloads:
                download_dict[file_type].append(url.replace("&amp;", "&"))

            record_id = re.search(r">(.*?)<", cells[2]).groups()[0]

            datum = {"id": cells[0], "record_id": record_id, "downloads": download_dict}
            index_data.append(datum)

        yield index_data


def get_responses_from_container_url(
    container_url: str, base_url: str, max_retries: int = 3
) -> Tuple[requests.Response, requests.Response, str]:
    """

    :param container_url:
    :param base_url:
    :param max_retries:
    :return:
    """
    session = requests.Session()

    retry = Retry(connect=max_retries, backoff_factor=0.5)
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)

    file_container_response = session.get(container_url)
    downloader._assert_status(file_container_response)

    file_container_response_text = file_container_response.text

    cookie = file_container_response.headers["Set-Cookie"].split(";")[0]

    file_container_frame_src = re.search(r"<FRAME SRC=\"/(.*?)\"", file_container_response_text).groups()[0]

    file_container_frame_response = session.get(base_url + file_container_frame_src, headers={"Cookie": cookie})
    downloader._assert_status(file_container_frame_response)
    file_container_frame_response_text = file_container_frame_response.text

    metadata_url, consent_url = re.search(
        r"setLabelMetadataStream\(.*, \"(.*?)\", .*, .*, .*, \"(.*?)\"", file_container_frame_response_text
    ).groups()

    metadata_response = requests.get(metadata_url)
    downloader._assert_status(metadata_response)

    # this returns either a consent reponse or directly the desired file content
    consent_response_or_file_response = session.get(consent_url, stream=True, headers={"Cookie": cookie})
    downloader._assert_status(consent_response_or_file_response)

    return metadata_response, consent_response_or_file_response, cookie


def get_metadata_from_response(metadata_response: requests.Response) -> dict:
    """

    :param metadata_response:
    :return:
    """
    metadata_text = metadata_response.text
    metadata_results = re.findall(r"<td>(.*?):</td>\s*<td>(.*?)<", metadata_text)
    metadata = {k: v for k, v in metadata_results}

    return metadata


@contextlib.contextmanager
def _stream_file_from_container_url(
    container_url: str, base_url: str, max_retries: int = 3, **kwargs: Any
) -> Iterator[Tuple[downloader.Response, Dict, Iterable[bytes]]]:
    """

    :param container_url:
    :param base_url:
    :param max_retries:
    :param kwargs:
    :return:
    """
    metadata_response, main_response, cookie = get_responses_from_container_url(
        container_url=container_url, base_url=base_url, max_retries=max_retries
    )

    metadata = get_metadata_from_response(metadata_response)

    main_response_text = main_response.text

    # search for copyrights uri in response
    copyrights_uri_search_results = re.search(r"copyrights_pid.src = \"/(.*?)\"", main_response_text)

    if copyrights_uri_search_results is None:
        # assume consent form did not appear
        yield (main_response, metadata, main_response.iter_content(chunk_size=io.DEFAULT_BUFFER_SIZE))
    else:
        # assume consent form did appear
        with requests.Session() as session:
            copyrights_uri = copyrights_uri_search_results.groups()[0]
            copyright_response = session.get(base_url + copyrights_uri, headers={"Cookie": cookie}, **kwargs)
            downloader._assert_status(copyright_response)

            download_url = re.search(r"window.location.href=\'(.*?)\'", main_response_text).groups()[0]

            with session.get(download_url, stream=True, headers={"Cookie": cookie}, **kwargs) as download_response:
                downloader._assert_status(download_response)

                yield (download_response, metadata, download_response.iter_content(chunk_size=io.DEFAULT_BUFFER_SIZE))


def _get_file_name_bsl_corpus(response: downloader.Response, metadata: Dict) -> str:
    """

    :param response:
    :param metadata:
    :return:
    """
    response_file_name = re.search(r"filename=%22(.*?)%22", response.headers["Content-Disposition"]).groups()[0]

    # return the response filename because in many cases the metadata filename is incorrect:
    # a) Identifier is not really the file name, especially if there are several ELAN files for a single identifier
    # b) instead of the identifier what is stored is the date, while in the "date" field some other information
    #    is stored, such as "Mini-DVD". Example: M20c

    return response_file_name


# noinspection PyProtectedMember
@utils.memoize()
def get_bsl_corpus_downloader(*args: Any, **kwargs: Any) -> "_BslCorpusDownloader":
    return _BslCorpusDownloader(*args, **kwargs)


# noinspection PyProtectedMember
class _BslCorpusDownloader(downloader._Downloader):
    def __init__(
        self, username: str, password: str, max_retries: int = 3, base_url: str = "http://digital-collections.ucl.ac.uk/", **kwargs
    ):
        super(_BslCorpusDownloader, self).__init__(**kwargs)

        self.username = username
        self.password = password
        self.max_retries = max_retries
        self.base_url = base_url

    def _sync_download(self, url: str, destination_path: str, verify: bool = True) -> downloader.DownloadResult:
        """Synchronous version of `download` method.
        To download through a proxy, the `HTTP_PROXY`, `HTTPS_PROXY`,
        `REQUESTS_CA_BUNDLE`,... environment variables can be exported, as
        described in:
        https://requests.readthedocs.io/en/master/user/advanced/#proxies
        Args:
          url: url to download
          destination_path: path where to write it
          verify: whether to verify ssl certificates
        Returns:
          None
        Raises:
          DownloadError: when download fails.
        """
        # files can have unknown names as a fallback

        with _stream_file_from_container_url(container_url=url, base_url=self.base_url, max_retries=self.max_retries, verify=verify) as (
            response,
            metadata,
            iter_content,
        ):
            # fname = downloader._get_filename(response)
            fname = _get_file_name_bsl_corpus(response, metadata)
            path = os.path.join(destination_path, fname)
            size = 0

            # Initialize the download size progress bar
            size_mb = 0
            unit_mb = units.MiB
            total_size = int(response.headers.get("Content-length", 0)) // unit_mb
            self._pbar_dl_size.update_total(total_size)
            with tf.io.gfile.GFile(path, "wb") as file_:
                checksum = self._checksumer_cls()
                for block in iter_content:
                    size += len(block)
                    checksum.update(block)
                    file_.write(block)

                    # Update the download size progress bar
                    size_mb += len(block)
                    if size_mb > unit_mb:
                        self._pbar_dl_size.update(size_mb // unit_mb)
                        size_mb %= unit_mb
        self._pbar_url.update(1)
        return downloader.DownloadResult(
            path=utils.as_path(path), url_info=checksums_lib.UrlInfo(checksum=checksum.hexdigest(), size=utils.Size(size), filename=fname,),
        )


class DownloadManagerBslCorpus(tfds.download.DownloadManager):
    def __init__(self, *, username: str, password: str, max_retries: int = 3, **kwargs):
        super().__init__(**kwargs)

        self.username = username
        self.password = password
        self.max_retries = max_retries

        self.__downloader = None

    @property
    def _downloader(self):
        if self.__downloader is None:
            self.__downloader = get_bsl_corpus_downloader(username=self.username, password=self.password, max_retries=self.max_retries)
        return self.__downloader


def get_elan_sentences_bsl_corpus(elan_path: str) -> Iterator:
    """
    Third release notes and annotation conventions:
    https://bslcorpusproject.org/wp-content/uploads/BSLCorpus_AnnotationConventions_v3.0_-March2017.pdf
    https://bslcorpusproject.org/wp-content/uploads/Notes-to-the-3rd-release-of-BSL-Corpus-annotations.pdf

    Names of tiers:

    - "LH-IDgloss" for left hand glosses
    - "RH-IDgloss" for right hand glosses
    - "Free Translation" for English translation

    :param elan_path:
    :return:
    """

    eaf = pympi.Elan.Eaf(elan_path)  # TODO add "suppress_version_warning=True" when pympi 1.7 is released

    timeslots = eaf.timeslots

    english_tier_name = "Free Translation"
    if english_tier_name not in eaf.tiers:
        return

    # tiers is defined as follows (http://dopefishh.github.io/pympi/Elan.html):
    #   {tier_name -> (aligned_annotations, reference_annotations, attributes, ordinal)}
    # aligned_annotations:
    #   [{id -> (begin_ts, end_ts, value, svg_ref)}]
    # reference_annotations:
    #   [{id -> (reference, value, previous, svg_ref)}]
    #
    # - "ts" means timeslot, which references a time value in miliseconds
    # - "value" is the actual annotation content
    # - "svg_ref" is an optional reference to an SVG image that is always None in our files

    english_text = list(eaf.tiers[english_tier_name][0].values())

    # collect all glosses in the entire ELAN file

    all_glosses = []

    for hand in ["R", "L"]:
        hand_tier = hand + "H-IDgloss"
        if hand_tier not in eaf.tiers:
            continue

        glosses = {}

        for gloss_id, (start, end, value, _) in eaf.tiers[hand_tier][0].items():
            glosses[gloss_id] = {"start": timeslots[start], "end": timeslots[end], "gloss": value, "hand": hand}

        all_glosses += list(glosses.values())

    for (start, end, value, _) in english_text:
        sentence = {"start": timeslots[start], "end": timeslots[end], "english": value}

        # Add glosses whose timestamps are within this sentence
        glosses_in_sentence = [item for item in all_glosses if item["start"] >= sentence["start"] and item["end"] <= sentence["end"]]

        sentence["glosses"] = list(sorted(glosses_in_sentence, key=lambda d: d["start"]))

        yield sentence


if __name__ == "__main__":
    # running this module as main writes around 6 *.eaf files to the current directory

    BSLCP_USERNAME = os.environ["BSLCP_USERNAME"]
    BSLCP_PASSWORD = os.environ["BSLCP_PASSWORD"]

    base_url = "http://digital-collections.ucl.ac.uk/"

    num_example_records = 20

    download_links_iterator = generate_download_links(
        username=BSLCP_USERNAME, password=BSLCP_PASSWORD, number_of_records=num_example_records
    )

    for data_per_results_page in download_links_iterator:

        for datum in data_per_results_page:
            # look for a record with an ELAN file
            if "EAF file" not in datum["downloads"].keys():
                continue

            resource_url_elan = datum["downloads"]["EAF file"][0]

            with _stream_file_from_container_url(container_url=resource_url_elan, base_url=base_url, max_retries=5) as (
                response,
                metadata,
                iter_content,
            ):

                file_name = _get_file_name_bsl_corpus(response, metadata)
                print("Found %s" % file_name)
                with open(file_name, "wb") as outfile:
                    for block in iter_content:
                        outfile.write(block)



"""Corpus NGT: an open access online corpus of movies with annotations of Sign Language of the Netherlands"""

import json
import re

import numpy as np
import tensorflow as tf
import tensorflow_datasets as tfds

from typing import Optional

from ..warning import dataset_warning
from ...datasets.config import SignDatasetConfig

_DESCRIPTION = """
An open access online corpus of movies with annotations of Sign Language of the Netherlands.
"""

_CITATION = """\
@inproceedings{dataset:Crasborn2008TheCN,
    title = {The Corpus NGT: An online corpus with annotations of Sign Language of the Netherlands 
    for professionals and laymen},
    author = {O. Crasborn and I. Zwitserlood},
    year = {2008}
}
"""

# This `ngt.json` file was created using `create_index.py`
INDEX_URL = "https://nlp.biu.ac.il/~amit/datasets/ngt.json"

# alternative mirror:
# INDEX_URL = "https://files.ifi.uzh.ch/cl/archiv/2022/easier/ngt.json"

_HOMEPAGE = "https://www.ru.nl/corpusngtuk/"

_VIDEO_RESOLUTION = (352, 288)

# the "combined view" of both participants is a wider video
_VIDEO_RESOLUTION_C = (704, 288)

_FRAMERATE = 25


def _get_speaker_id_from_filename(filename: str) -> Optional[str]:
    """

    :param filename:
    :return:
    """
    matches = re.findall(r"S\d+", filename)
    if len(matches) == 0:
        return None

    return matches[0]


def _get_view_to_speaker_mapping(datum_dict: dict) -> dict:
    """

    :param datum_dict:
    :return:
    """
    speaker_set = set()
    for key in datum_dict.keys():
        speaker_id = _get_speaker_id_from_filename(key)
        if speaker_id is None:
            continue
        speaker_set.add(speaker_id)
    speakers = list(sorted(speaker_set))

    assert 0 < len(speakers) <= 2, "Corpus example has 0 or more than 2 speakers: %s" % str(datum_dict)

    return {"a": speakers[0], "b": speakers[1] if len(speakers) == 2 else None}


class NGTCorpus(tfds.core.GeneratorBasedBuilder):
    """
    DatasetBuilder for NGT corpus dataset.

    Design decisions about camera views and speaker IDs:

    There are 4 different camera views:
    - "f": face of speaker, frontal view, resolution: 352 x 288
    - "b": body of speaker, including the face, frontal view, resolution: 352 x 288
    - "t": top, the speaker filmed from above, resolution: 352 x 288
    - no view identifier and no speaker (i.e. "CNGT0100.mpg" for example): wider-angle video that
      shows the entire scene, resolution: 704 x 288

    For this loader we decided to:
    - ignore the "t" (top) and "f" (face) views
    - use "a" and "b" as the video feature identifiers for the "b" (body) view of the first and second speaker
      in each corpus example, ignoring the fact that every speaker has a unique ID
    - use "c" as the video feature identifier for the combined view
    """

    VERSION = tfds.core.Version("3.0.0")
    RELEASE_NOTES = {
        "3.0.0": "3rd release",
    }

    BUILDER_CONFIGS = [
        SignDatasetConfig(name="default", include_video=True, include_pose=None),
        SignDatasetConfig(name="videos", include_video=True, include_pose=None),
        SignDatasetConfig(name="annotations", include_video=False, include_pose=None),
    ]

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""
        features = {
            "id": tfds.features.Text(),
            "paths": {"eaf": tfds.features.Text(),},
        }

        if self._builder_config.include_video:
            features["fps"] = tf.int32
            video_ids = ["a", "b", "c"]
            if self._builder_config.process_video:
                features["videos"] = {
                    "a": self._builder_config.video_feature(_VIDEO_RESOLUTION),
                    "b": self._builder_config.video_feature(_VIDEO_RESOLUTION),
                    "c": self._builder_config.video_feature(_VIDEO_RESOLUTION_C),
                }
            features["paths"]["videos"] = {_id: tfds.features.Text() for _id in video_ids}

        # Add poses if requested
        if self._builder_config.include_pose is not None:
            raise NotImplementedError("Poses are currently not available for the NGT corpus.")

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features),
            homepage=_HOMEPAGE,
            supervised_keys=None,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        dataset_warning(self)

        index_path = dl_manager.download(INDEX_URL)

        with open(index_path, "r", encoding="utf-8") as f:
            index_data = json.load(f)

        # remove corpus examples if videos are restricted (= only an empty ELAN file is available, or only the "c"
        # or "t" video without annotations)
        keys_to_be_removed = set()
        for datum_key, datum in index_data.items():
            if len(datum.keys()) == 1:
                keys_to_be_removed.add(datum_key)

        # this is currently removing the following keys:
        # {'CNGT0025', 'CNGT0876', 'CNGT1542', 'CNGT0020', 'CNGT0314'}

        for key in keys_to_be_removed:
            del index_data[key]

        if self._builder_config.include_video:
            # rename "body" video keys, ignore other views
            for datum in index_data.values():

                # determine which video URL corresponds to abstract views "a" and "b"
                view_to_speaker_mapping = _get_view_to_speaker_mapping(datum)

                speaker_a = view_to_speaker_mapping["a"]
                video_key_a = "video_" + speaker_a + "_b"  # this "_b" is for "body" view
                datum["video_a"] = datum[video_key_a]

                speaker_b = view_to_speaker_mapping["b"]
                if speaker_b is not None:
                    video_key_b = "video_" + speaker_b + "_b"  # this "_b" is for "body" view
                    datum["video_b"] = datum[video_key_b]

                video_keys_to_keep = ("video_a", "video_b", "video_c")
                all_keys = list(datum.keys())

                for key in all_keys:
                    if key.startswith("video"):
                        if key not in video_keys_to_keep:
                            del datum[key]
        else:
            # Don't download videos if not necessary
            for datum in index_data.values():
                keys = list(datum.keys())
                for key in keys:
                    if key.startswith("video"):
                        del datum[key]

        # never download audio files
        for datum in index_data.values():
            if "audio" in datum.keys():
                del datum["audio"]

        urls = {url: url for datum in index_data.values() for url in datum.values() if url is not None}

        local_paths = dl_manager.download(urls)

        processed_data = {
            _id: {k: local_paths[v] if v is not None else None for k, v in datum.items()} for _id, datum in index_data.items()
        }

        return [tfds.core.SplitGenerator(name=tfds.Split.TRAIN, gen_kwargs={"data": processed_data})]

    def _generate_examples(self, data):
        """ Yields examples. """

        default_fps = _FRAMERATE
        default_video = np.zeros((0, 0, 0, 3))  # Empty video

        for _id, datum in list(data.items()):
            features = {
                "id": _id,
                "paths": {t: str(datum[t]) if t in datum else "" for t in ["eaf"]},
            }

            if self._builder_config.include_video:
                videos = {t: str(datum["video_" + t]) if ("video_" + t) in datum else "" for t in ["a", "b", "c"]}

                features["fps"] = self._builder_config.fps if self._builder_config.fps is not None else default_fps
                features["paths"]["videos"] = videos
                if self._builder_config.process_video:
                    features["videos"] = {t: v if v != "" else default_video for t, v in videos.items()}

            if self._builder_config.include_pose == "openpose":
                raise NotImplementedError("OpenPose poses are currently not available for the NGT corpus.")

            if self._builder_config.include_pose == "holistic":
                raise NotImplementedError("Holistic poses are currently not available for the NGT corpus.")

            yield _id, features



"""A German Sign Language (DGS) lexicon for natural science subjects."""
import json
import os

import tensorflow as tf
import tensorflow_datasets as tfds

from ..warning import dataset_warning
from ...datasets import SignDatasetConfig
import urllib.request
import cv2

from ...utils.signwriting.ocr import image_to_fsw

_DESCRIPTION = """
The specialist signs developed in the project break down barriers for deaf people 
and thus facilitate their access to education in natural science subjects. 
Improved communication enables deaf people in the MINT subjects to participate better in research and science, 
from school to university to independent work. With the MINT specialist sign language dictionary, 
you have a better chance of developing your scientific ideas as a researcher; 
it makes work in the laboratory easier and enriches your own career planning, as research results, 
for example, can be better communicated. For the first time, teachers, 
students and pupils have access to a uniform and constantly expanding MINT specialist sign language dictionary,
"""

# TODO(sign2mint): BibTeX citation
_CITATION = """
"""

OCR_CACHE_PATH = os.path.join(os.path.dirname(__file__), "ocr_cache.txt")


class Sign2MINT(tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for sign2mint dataset."""

    VERSION = tfds.core.Version("1.0.0")
    RELEASE_NOTES = {
        "1.0.0": "Initial release.",
    }

    BUILDER_CONFIGS = [
        SignDatasetConfig(name="default", include_video=True),
        SignDatasetConfig(name="annotations", include_video=False),
    ]

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""

        features = {
            "id": tfds.features.Text(),
            "fachbegriff": tfds.features.Text(),
            "fachgebiete": tfds.features.Sequence(tfds.features.Text()),
            "ursprung": tfds.features.Sequence(tfds.features.Text()),
            "verwendungskontext": tfds.features.Sequence(tfds.features.Text()),
            "definition": tfds.features.Text(),
            "bedeutungsnummern": tfds.features.Text(),
            "wortlink": tfds.features.Text(),
            "wikipedialink": tfds.features.Text(),
            "otherlink": tfds.features.Text(),
            "variants": tf.int32,
            "gebaerdenschrift": {
                "symbolIds": tfds.features.Sequence(tfds.features.Text()),
            },
            "swu": tfds.features.Text(),
        }

        if self._builder_config.include_video and self._builder_config.process_video:
            features["video"] = self._builder_config.video_feature((1920, 1080))
        else:
            features["video"] = tfds.features.Text()

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features),
            homepage="https://sign2mint.de/",
            supervised_keys=None,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        dataset_warning(self)

        annotations_path = dl_manager.download("https://sign2mint.de/api/entries/all/")

        local_videos = {}
        if self._builder_config.include_video and self._builder_config.process_video:
            with open(annotations_path, "r", encoding="utf-8") as f:
                annotations = json.load(f)
                videos = [a["videoLink"] for a in annotations]
                video_paths = dl_manager.download(videos)
                local_videos = {k: v for k, v in zip(videos, video_paths)}

        return {"train": self._generate_examples(annotations_path, local_videos)}

    ocr_cache = None

    def _ocr(self, datum):
        # sign2mint used to include only a photo of signwriting, which required OCR to get the FSW
        if self.ocr_cache is None:
            with open(OCR_CACHE_PATH, "r") as f:
                lines = [l.split(" ") for l in f.readlines()]
                self.ocr_cache = {l[0]: l[1] for l in lines}

        image_url = datum["gebaerdenschrift"]["url"].replace("&transparent=true", "")
        if image_url in self.ocr_cache:
            return self.ocr_cache[image_url]

        urllib.request.urlretrieve(image_url, "sign.png")
        img_rgb = cv2.imread("sign.png")

        symbols = datum["gebaerdenschrift"]["symbolIds"]

        fsw = image_to_fsw(img_rgb, symbols)

        with open(OCR_CACHE_PATH, "a") as f:
            f.write(image_url + " " + fsw + "\n")

        return fsw

    def _generate_examples(self, annotations_path, local_videos):
        """Yields examples."""

        with open(annotations_path, "r", encoding="utf-8") as f:
            data = json.load(f)
            for datum in data:
                del datum["empfehlung"]  # remove unused property

                video_link = datum["videoLink"]
                del datum["videoLink"]
                del datum["videoThumbnailLink"]

                datum["video"] = local_videos[video_link] if video_link in local_videos else video_link
                datum["gebaerdenschrift"]["symbolIds"] = [
                    s["symbolKey"] for s in datum["gebaerdenschrift"]["symbolIds"] if s["symbolKey"] != ""
                ]
                # datum["gebaerdenschrift"]["fsw"] = self._ocr(datum)

                yield datum["id"], datum



"""How2Sign: A multimodal and multiview continuous American Sign Language (ASL) dataset"""
import csv
import tarfile
from os import path

import tensorflow as tf
import tensorflow_datasets as tfds
from tensorflow.io.gfile import GFile

from ..warning import dataset_warning
from ...datasets.config import SignDatasetConfig

_DESCRIPTION = """
ChicagoFSWild+ contains 55,232 fingerspelling sequences signed by 260 signers.
"""

_CITATION = """
@article{fs18iccv,
author = {B. Shi, A. Martinez Del Rio, J. Keane, D. Brentari, G. Shakhnarovich, and K. Livescu},
title = {Fingerspelling recognition in the wild with iterative visual attention},
journal = {ICCV},
year = {2019},
month = {October}
}
@article{fs18slt,
author = {B. Shi, A. Martinez Del Rio, J. Keane, J. Michaux, D. Brentari, G. Shakhnarovich, and K. Livescu},
title = {American Sign Language fingerspelling recognition in the wild},
journal = {SLT},
year = {2018},
month = {December}
}
"""

_VERSIONS = {
    "2.0.0": {"name": "ChicagoFSWildPlus", "url": "https://dl.ttic.edu/ChicagoFSWildPlus.tgz"},
    "1.0.0": {"name": "ChicagoFSWild", "url": "https://dl.ttic.edu/ChicagoFSWild.tgz"},
}

_CHICAGO_FS_WILD_PLUS_URL = "https://dl.ttic.edu/ChicagoFSWildPlus.tgz"
_CHICAGO_FS_WILD_URL = "https://dl.ttic.edu/ChicagoFSWild.tgz"


class ChicagoFSWild(tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for how2sign dataset."""

    VERSION = tfds.core.Version("2.0.0")
    RELEASE_NOTES = {k: v["name"] for k, v in _VERSIONS.items()}

    BUILDER_CONFIGS = [SignDatasetConfig(name="default", include_video=True)]

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""

        features = {
            "id": tfds.features.Text(),
            "text": tfds.features.Text(),
            "description": tfds.features.Text(),
            "video_url": tfds.features.Text(),
            "start": tfds.features.Text(),
            "signer": tfds.features.Text(),
            "metadata": {"frames": tf.int32, "width": tf.int32, "height": tf.int32,},
        }

        if self._builder_config.include_video:
            features["video"] = self._builder_config.video_feature((480, 360))

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features),
            homepage="https://home.ttic.edu/~klivescu/ChicagoFSWild.htm",
            supervised_keys=None,
            citation=_CITATION,
        )

    def _version_details(self, key: str):
        return _VERSIONS[str(self.version)][key]

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        """Returns SplitGenerators."""
        dataset_warning(self)

        archive = dl_manager.download_and_extract(self._version_details("url"))

        return [
            tfds.core.SplitGenerator(name=tfds.Split.TRAIN, gen_kwargs={"archive_path": archive, "split": "train"}),
            tfds.core.SplitGenerator(name=tfds.Split.VALIDATION, gen_kwargs={"archive_path": archive, "split": "dev"}),
            tfds.core.SplitGenerator(name=tfds.Split.TEST, gen_kwargs={"archive_path": archive, "split": "test"}),
        ]

    def _generate_examples(self, archive_path: str, split: str):
        """ Yields examples. """

        v_name = self._version_details("name")
        archive_directory = path.join(archive_path, v_name) if v_name == "ChicagoFSWild" else archive_path

        frames_directory = path.join(archive_directory, "frames")
        frames_directory = path.join(frames_directory, v_name) if v_name == "ChicagoFSWildPlus" else frames_directory

        if not path.exists(frames_directory):
            print("Extracting Frames Archive")
            tar = tarfile.open(path.join(archive_directory, v_name + "-Frames.tgz"))
            tar.extractall(path=frames_directory)
            tar.close()

        with GFile(path.join(archive_directory, v_name + ".csv"), "r") as csv_file:
            csv_data = csv.reader(csv_file, delimiter=",")
            next(csv_data)  # Ignore the header

            for row in csv_data:
                if row[10] == split:
                    _id = row[1].replace("/", "-").replace("_(youtube)", "").replace("_(nad)", "")

                    datum = {
                        "id": _id,
                        "text": row[7],
                        "description": row[9],
                        "video_url": row[2],
                        "start": row[3],
                        "signer": row[-1],
                        "metadata": {"frames": int(row[4]), "width": int(row[5]), "height": int(row[6])},
                    }

                    if self._builder_config.include_video:
                        frames_base = path.join(frames_directory, row[1])

                        datum["video"] = [path.join(frames_base, name) for name in sorted(tf.io.gfile.listdir(frames_base))]

                    yield _id, datum



"""SignTyp dataset."""
import re
from datetime import datetime

import requests
import tensorflow as tf
import tensorflow_datasets as tfds

from ..warning import dataset_warning
from ...datasets import SignDatasetConfig

_DESCRIPTION = """
SignTyp is a work in progress. We currently have over 20,000 videos for signers from Brazil, China, Croatia, Denmark, Egypt, Estonia, Ethiopia, France, Germany, Haiti, Honduras, Iraq, Italy,Lithuania, Nepal, Nicaragua, the Philippines, Portugal, Russia, Rwanda, Sri Lanka, Turkey, Uganda, the United States and Venezuela. We have complete sets of approximately 1000 video recordings each for China, Denmark, Egypt, Ethiopia, Haiti, Honduras, Iraq, Lithuania, Portugal,the Phillipines, Russia (2 signers), Turkey, and the United States. Clipping has been completed for all sets except a second Russian signer.

We currently have SignWriting transcriptions for approximately 9,000 signs, with more in progress. Each transcription includes detailed information on handshapes, orientation, actions, locations, as well as special sequencing and articulator relationship information. There is a downloadable file for all transcribed information for sign language researchers.

SignWriting transcription can be seen in our SignPuddles - see link below.

The project is sponsored by NSF (National Science Foundation) and is made possible by grant BCS-1049510 to the University of Connecticut and Harry van der Hulst.
"""

# TODO(SignTyp): BibTeX citation
_CITATION = """
"""


class SignTyp(tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for SignTyp dataset."""

    VERSION = tfds.core.Version("1.0.0")
    RELEASE_NOTES = {
        "1.0.0": "Initial release.",
    }

    BUILDER_CONFIGS = [
        SignDatasetConfig(name="default", include_video=True, process_video=False, extra={"PHPSESSID": "hj9co07ct7f5noq529no9u09l4"})
    ]

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""

        features = {
            "id": tf.int32,
            "created_date": tfds.features.Text(),
            "modified_date": tfds.features.Text(),
            "sign_writing": tfds.features.Text(),
            "sign_language": tfds.features.Text(),
            "video": tfds.features.Text(),
        }

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features),
            homepage="https://signtyp.uconn.edu/",
            supervised_keys=None,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        """Returns SplitGenerators."""
        dataset_warning(self)

        if "PHPSESSID" not in self._builder_config.extra:
            raise Exception(
                "Missing PHPSESSID extra parameter. Go to https://signtyp.uconn.edu/signpuddle/ and copy your PHPSESSID from any network request."
            )

        cookies = {"PHPSESSID": self._builder_config.extra["PHPSESSID"]}

        headers = {
            "Connection": "keep-alive",
            "Cache-Control": "max-age=0",
            "sec-ch-ua": '" Not A;Brand";v="99", "Chromium";v="96", "Google Chrome";v="96"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "Upgrade-Insecure-Requests": "1",
            "Origin": "https://signtyp.uconn.edu",
            "Content-Type": "application/x-www-form-urlencoded",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.45 Safari/537.36",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            "Sec-Fetch-Site": "same-origin",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-User": "?1",
            "Sec-Fetch-Dest": "document",
            "Referer": "https://signtyp.uconn.edu/signpuddle/export.php?ui=1&sgn=9032",
            "Accept-Language": "en-US,en;q=0.9,he;q=0.8",
        }

        data = {"ex_source": "All", "action": "Download"}

        res = requests.post("https://signtyp.uconn.edu/signpuddle/export.php", data=data, headers=headers, cookies=cookies)

        spml = res.text

        if not spml.startswith('<?xml version="1.0" encoding="UTF-8"?>'):
            raise Exception("PHPSESSID might be expired.")

        return {
            "train": self._generate_examples(spml),
        }

    def _generate_examples(self, spml):
        regex = '<entry id="(.*?)".*?cdt="(.*?)" mdt="(.*?)".*?term>(.*?)<.*?lxsg: (.*?)].*(http.*?)"'

        for row in spml.splitlines():
            datum = re.search(regex, row)

            if datum is not None:
                _id = datum.group(1)

                yield _id, {
                    "id": int(_id),
                    "created_date": str(datetime.fromtimestamp(int(datum.group(2)))),
                    "modified_date": str(datetime.fromtimestamp(int(datum.group(3)))),
                    "sign_writing": datum.group(4),
                    "sign_language": datum.group(5),
                    "video": datum.group(6),
                }



import os
from typing import Optional

import numpy as np
import numpy.ma as ma
import tensorflow as tf
from pose_format import Pose, PoseHeader
from pose_format.numpy.pose_body import NumPyPoseBody
from pose_format.utils.reader import BufferReader
from tensorflow_datasets.core.features import feature
from tensorflow_datasets.core.features.feature import Documentation
from tensorflow_datasets.core.utils import Json
from etils import epath


def read_header(header_path: str):
    with open(header_path, "rb") as f:
        reader = BufferReader(f.read())
        header = PoseHeader.read(reader)
        return header, reader.read_offset


def read_body(pose_bytes: bytes, header: PoseHeader, read_offset: Optional[int] = None):
    reader = BufferReader(pose_bytes)
    if header is None:
        header = PoseHeader.read(reader)
    else:
        reader.read_offset = read_offset

    return NumPyPoseBody.read(header, reader)


# TODO write doc


class PoseFeature(feature.FeatureConnector):
    """`FeatureConnector` for poses.

    During `_generate_examples`, the feature connector accept as input any of:

      * `str`: path to a {bmp,gif,jpeg,png} image (ex: `/path/to/img.png`).
      * `np.array`: 3d `np.uint8` array representing an image.
      * A file object containing the png or jpeg encoded image string (ex:
        `io.BytesIO(encoded_img_bytes)`)

    Output:

      `tf.Tensor` of type `tf.uint8` and shape `[height, width, num_channels]`
      for BMP, JPEG, and PNG images and shape `[num_frames, height, width, 3]` for
      GIF images.

    Example:

      * In the `tfds.core.DatasetInfo` object:

      ```python
      features=features.FeaturesDict({
          "input": features.Image(),
          "target": features.Image(shape=(None, None, 1),
                                     encoding_format="png"),
      })
      ```

      * During generation:

      ```python
      yield {
          "input": "path/to/img.jpg",
          "target": np.ones(shape=(64, 64, 1), dtype=np.uint8),
      }
      ```
    """

    def __init__(self, *, shape=None, header_path: str = None, encoding_format: str = None, stride: int = 1,
                 dtype=tf.float32):
        """Construct the connector.

        Args:
          shape: tuple of ints or None, the shape of decoded image.
            For GIF images: (num_frames, height, width, channels=3). num_frames,
              height and width can be None.
            For other images: (height, width, channels). height and width can be
              None. See `tf.image.encode_*` for doc on channels parameter.
            Defaults to (None, None, 3).
          dtype: tf.uint16 or tf.uint8 (default).
            tf.uint16 can be used only with png encoding_format
          encoding_format: "jpeg" or "png". Format to serialize `np.ndarray` images
            on disk. If None, encode images as PNG.
            If image is loaded from {bmg,gif,jpeg,png} file, this parameter is
            ignored, and file original encoding is used.

        Raises:
          ValueError: If the shape is invalid
        """
        # Set and validate values
        self._shape = shape or (None, None, None, 3)
        self._dtype = dtype
        self._encoding_format = encoding_format or "pose"

        self._doc = Documentation()

        assert int(stride) == stride, "Video fps must be divisible by custom fps, when loading poses"
        self.stride = int(stride)

        if header_path is not None:
            self._header, self._read_offset = read_header(header_path)
        else:
            self._header, self._read_offset = None, None

        self._runner = None

    def get_tensor_info(self):
        # Image is returned as a 3-d uint8 tf.Tensor.
        conf_shape = tuple(list(self._shape)[:3])
        return {
            "data": feature.TensorInfo(shape=self._shape, dtype=self._dtype),
            "conf": feature.TensorInfo(shape=conf_shape, dtype=self._dtype),
            "fps": feature.TensorInfo(shape=(), dtype=tf.int32),
        }

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_runner"] = None
        return state

    def encode_body(self, body: NumPyPoseBody):
        # print("shape", body.data.shape, np.isfinite(body.data).all())

        if self.stride != 1:
            body = body.slice_step(self.stride)

        data = body.data.data
        confidence = body.confidence

        return {"data": data, "conf": confidence, "fps": int(body.fps)}

    def encode_example(self, pose_path_or_fobj):
        """Convert the given image into a dict convertible to tf example."""

        if pose_path_or_fobj is None:
            # Create 0 size tensors
            data_shape = list(self._shape)
            data_shape[0] = 0
            conf_shape = tuple(data_shape[:3])
            data_shape = tuple(data_shape)

            pose_body = NumPyPoseBody(data=ma.zeros(data_shape), confidence=np.zeros(conf_shape), fps=0)
            return self.encode_body(pose_body)
        elif isinstance(pose_path_or_fobj, Pose):
            return self.encode_body(pose_path_or_fobj.body)
        elif isinstance(pose_path_or_fobj, epath.PathLikeCls):
            pose_path_or_fobj = os.fspath(pose_path_or_fobj)
            with tf.io.gfile.GFile(pose_path_or_fobj, "rb") as pose_f:
                encoded_pose = pose_f.read()
        elif isinstance(pose_path_or_fobj, bytes):
            encoded_pose = pose_path_or_fobj
        else:
            encoded_pose = pose_path_or_fobj.read()

        if self._encoding_format == "pose":
            pose_body = read_body(encoded_pose, self._header, self._read_offset)
            return self.encode_body(pose_body)
        else:
            raise Exception("Unknown encoding format '%s'" % self._encoding_format)

    # def decode_example(self, example):
    #     """Reconstruct the image from the tf example."""
    #     print("Decoding", example)
    #     print(example["data"].shape)
    #     print(example["data"])
    #     raise Exception()
    #     img = tf.image.decode_image(
    #         example, channels=self._shape[-1], dtype=self._dtype)
    #     img.set_shape(self._shape)
    #     return img

    @classmethod
    def from_json_content(cls, value: Json) -> "Pose":
        shape = tuple(value["shape"])
        encoding_format = value["encoding_format"]
        return cls(shape=shape, encoding_format=encoding_format)

    def to_json_content(self) -> Json:
        print("to_json_content", {"shape": list(self._shape), "encoding_format": self._encoding_format, })
        return {
            "shape": list(self._shape),
            "encoding_format": self._encoding_format,
        }



from typing import List, Optional, Tuple

import cv2
import tensorflow_datasets as tfds


class SignDatasetConfig(tfds.core.BuilderConfig):
    """General BuilderConfig for sign language datasets."""

    def __init__(
        self,
        include_video: bool = True,
        process_video: bool = None,
        include_pose: Optional[str] = None,
        fps: Optional[float] = None,
        resolution: Optional[Tuple[int, int]] = None,
        extra: dict = {},
        **kwargs,
    ):
        """Constructs a RWTHPhoenix2014TConfig.
        Args:
          include_video: bool, whether to include videos in the data.
          process_video: bool, whether to load the videos as tensors or paths
          include_pose: str, what pose data to include.
          fps: float, what pose data to include.
          resolution: (int, int), what resolution of videos to load.
          split: specify a known split identifier (optional)
          **kwargs: keyword arguments forwarded to super.
        """
        super(SignDatasetConfig, self).__init__(**kwargs)
        self.include_video = include_video
        self.process_video = process_video if process_video is not None else include_video
        self.include_pose = include_pose.lower() if include_pose is not None else None

        self.fps = fps
        self.resolution = resolution
        self.extra = extra

    def ffmpeg_args(self):
        args: List[str] = []

        if self.fps is not None:
            args += ["-r", str(self.fps)]

        if self.resolution is not None:
            w, h = self.resolution
            ratio = str(w) + ":" + str(h)
            scale = ratio + ":force_original_aspect_ratio=decrease"
            pad = ratio + ":(ow-iw)/2:(oh-ih)/2"
            args += ["-vf", "scale=" + scale + ",pad=" + pad + ",setsar=1"]

        return args

    def video_feature(self, default_resolution: Tuple[int, int], channels=3):
        w, h = self.resolution if self.resolution is not None else default_resolution

        ffmpeg_extra_args = self.ffmpeg_args()
        return VideoFeature(shape=(None, h, w, channels), ffmpeg_extra_args=ffmpeg_extra_args)


class VideoFeature(tfds.features.Video):
    def encode_example(self, video_or_path_or_fobj):
        """TFDS does not scale images, so this class handles it"""

        # Load images if list of file paths
        if isinstance(video_or_path_or_fobj, list) and isinstance(video_or_path_or_fobj[0], str):
            _, h, w, _ = self.shape
            video_or_path_or_fobj = [cv2.resize(cv2.imread(f), (w, h)) for f in video_or_path_or_fobj]

        # In case where additional ffmpeg parameters are needed
        if isinstance(video_or_path_or_fobj, dict) and "video" in video_or_path_or_fobj and isinstance(video_or_path_or_fobj["video"], str):
            old_args = list(self._extra_ffmpeg_args)
            self._extra_ffmpeg_args += video_or_path_or_fobj["ffmpeg_args"]
            result = super(VideoFeature, self).encode_example(video_or_path_or_fobj["video"])
            self._extra_ffmpeg_args = old_args
            return result

        return super(VideoFeature, self).encode_example(video_or_path_or_fobj)



import urllib.request
from urllib.error import ContentTooShortError

_ASLPRO_HEADERS = [
    ("User-Agent", "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.9.0.7) Gecko/2009021910 Firefox/3.0.7"),
    ("Referer", "http://www.aslpro.com/cgi-bin/aslpro/aslpro.cgi"),
]


def download_aslpro(url: str, dst_path: str):
    # ASL Pro videos are in swf format
    opener = urllib.request.build_opener()
    opener.addheaders = _ASLPRO_HEADERS
    urllib.request.install_opener(opener)

    try:
        urllib.request.urlretrieve(url, dst_path)
    except ContentTooShortError:
        pass



"""A list of 1000+ concepts with a sign language equivalent in each of the four project languages."""
from os import path
import tensorflow_datasets as tfds
import re

from ..warning import dataset_warning
from ...datasets import SignDatasetConfig
from ...utils.features import PoseFeature

_DESCRIPTION = """
A list of 1000+ concepts with a sign language equivalent in each of the four project languages
"""

_CITATION = """
@inproceedings{efthimiou2010dicta,
  title={Dicta-sign--sign language recognition, generation and modelling: a research effort with applications in deaf communication},
  author={Efthimiou, Eleni and Fontinea, Stavroula-Evita and Hanke, Thomas and Glauert, John and Bowden, Rihard and Braffort, Annelies and Collet, Christophe and Maragos, Petros and Goudenove, Fran{\c{c}}ois},
  booktitle={Proceedings of the 4th Workshop on the Representation and Processing of Sign Languages: Corpora and Sign Language Technologies},
  pages={80--83},
  year={2010}
}
"""

SPOKEN_LANGUAGES = {"BSL": "en", "DGS": "de", "LSF": "fr", "GSL": "el"}

_POSE_URLS = {"holistic": "https://nlp.biu.ac.il/~amit/datasets/poses/holistic/dicta_sign.tar.gz"}
_POSE_HEADERS = {"holistic": path.join(path.dirname(path.realpath(__file__)), "holistic.header")}


class DictaSign(tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for sign2mint dataset."""

    VERSION = tfds.core.Version("1.0.0")
    RELEASE_NOTES = {
        "1.0.0": "Initial release.",
    }

    BUILDER_CONFIGS = [
        SignDatasetConfig(name="default", include_video=True, include_pose="holistic"),
        SignDatasetConfig(name="poses", include_video=False, include_pose="holistic"),
        SignDatasetConfig(name="annotations", include_video=False),
    ]

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""

        features = {
            "id": tfds.features.Text(),
            "signed_language": tfds.features.Text(),
            "spoken_language": tfds.features.Text(),
            "text": tfds.features.Text(),
            "gloss": tfds.features.Text(),
            "hamnosys": tfds.features.Text(),
        }

        resolution = (320, 240)
        if self._builder_config.resolution is None:
            # Required in order to pad all videos equally
            self._builder_config.resolution = resolution

        if self._builder_config.include_video and self._builder_config.process_video:
            features["video"] = self._builder_config.video_feature(resolution)
        else:
            features["video"] = tfds.features.Text()

        if self._builder_config.include_pose == "holistic":
            pose_header_path = _POSE_HEADERS[self._builder_config.include_pose]
            stride = 1 if self._builder_config.fps is None else 25 / self._builder_config.fps
            features["pose"] = PoseFeature(shape=(None, 1, 543, 3), header_path=pose_header_path, stride=stride)

        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(features),
            homepage="https://www.sign-lang.uni-hamburg.de/dicta-sign/portal/",
            supervised_keys=None,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        dataset_warning(self)

        concepts_path = dl_manager.download("https://www.sign-lang.uni-hamburg.de/dicta-sign/portal/concepts/concepts_eng.html")

        if self._builder_config.include_pose is not None:
            poses_path = dl_manager.download_and_extract(_POSE_URLS[self._builder_config.include_pose])
        else:
            poses_path = None

        regex = r"<a href=\"cs\/cs_(\d*)\.html\"><img"
        with open(concepts_path, "r", encoding="utf-8") as concepts_f:
            matches = re.finditer(regex, concepts_f.read(), re.MULTILINE)
            concept_url = "https://www.sign-lang.uni-hamburg.de/dicta-sign/portal/concepts/cs/cs_{}.html"
            concept_urls = [concept_url.format(match.groups()[0]) for match in matches]

        concept_paths = dl_manager.download(concept_urls)

        # These video suffixes are not found
        filter_videos = ["na.webm", "bsl/293.webm", "bsl/596.webm", "gsl/998.webm", "gsl/999.webm"]

        regex = r"\"title-c\">(.*?)<[\s\S]*?title-l1\">(.*?)&[\s\S]*?<source src=\".\/..\/(.*?)\" type=\"video\/webm[\s\S]*?gloss\">(.*?)<\/div[\s\S]*?\"hns\">(.*?)<"
        concepts = []
        for concept_path in concept_paths:
            with open(concept_path, "r", encoding="utf-8") as concept_f:
                matches = re.findall(regex, concept_f.read(), re.MULTILINE)

                assert len(matches) == 4, "Concept does not include 4 matches {}".format(concept_path)

                for match in matches:
                    concept = {
                        "id": match[2].split("/")[1].split(".")[0] + "_" + match[1],
                        "signed_language": match[1],
                        "spoken_language": SPOKEN_LANGUAGES[match[1]],
                        "text": match[0],
                        "gloss": match[3],
                        "hamnosys": match[4],
                        "video": "https://www.sign-lang.uni-hamburg.de/dicta-sign/portal/concepts/" + match[2],
                    }
                    if not any(map(lambda t: concept["video"].endswith(t), filter_videos)):
                        concepts.append(concept)

        if self._builder_config.include_video:
            video_urls = [c["video"] for c in concepts]
            video_paths = dl_manager.download(video_urls)

            for video_path, concept in zip(video_paths, concepts):
                concept["video"] = video_path if self._builder_config.process_video else str(video_path)

        return {"train": self._generate_examples(concepts, poses_path)}

    def _generate_examples(self, concepts, poses_path: str):
        """Yields examples."""

        for concept in concepts:
            if poses_path is not None:
                concept["pose"] = path.join(poses_path, "dicta_sign", concept["id"] + ".pose")

            yield concept["id"], concept



"""ngt_corpus dataset tests."""

import itertools

import tensorflow_datasets as tfds

from unittest import TestCase

from ... import datasets
from ...datasets.config import SignDatasetConfig
from ...datasets.ngt_corpus.ngt_corpus_utils import get_elan_sentences_ngt_corpus


class TestNgtCorpus(TestCase):
    def test_ngt_corpus_loader(self):

        config = SignDatasetConfig(name="only-annotations", version="1.0.0", include_video=False)
        ngt = tfds.load(name="ngt_corpus", builder_kwargs={"config": config})

        for datum in itertools.islice(ngt["train"], 0, 10):
            print(datum)


class TestNgtCorpusUtils(TestCase):
    def test_ngt_get_elan_sentences(self):

        config = SignDatasetConfig(name="only-annotations", version="1.0.0", include_video=False)
        ngt = tfds.load(name="ngt_corpus", builder_kwargs={"config": config})

        for datum in itertools.islice(ngt["train"], 0, 50):
            print(datum["id"].numpy().decode("utf-8"))
            elan_path = datum["paths"]["eaf"].numpy().decode("utf-8")

            sentences = get_elan_sentences_ngt_corpus(elan_path)

            for sentence in sentences:
                print(sentence)



"""
Utils file contains a subset of utils from:
https://github.com/bricksdont/sign-sockeye-baselines/blob/f056c50765bf5f088419591b505b9f967605d483/scripts/preprocessing/convert_and_split_data.py
"""

import os
import datetime
from typing import Dict

import cv2
import tarfile
import tempfile

import numpy as np

from pose_format import Pose, PoseHeader
from pose_format.numpy import NumPyPoseBody
from pose_format.pose_header import PoseHeaderDimensions
from pose_format.utils.openpose_135 import load_openpose_135_directory
from pose_format.utils.openpose import load_frames_directory_dict


def extract_tar_xz_file(filepath: str, target_dir: str):
    """

    :param filepath:
    :param target_dir:
    :return:
    """
    with tarfile.open(filepath) as tar_handle:
        def is_within_directory(directory, target):
            
            abs_directory = os.path.abspath(directory)
            abs_target = os.path.abspath(target)
        
            prefix = os.path.commonprefix([abs_directory, abs_target])
            
            return prefix == abs_directory
        
        def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
        
            for member in tar.getmembers():
                member_path = os.path.join(path, member.name)
                if not is_within_directory(path, member_path):
                    raise Exception("Attempted Path Traversal in Tar File")
        
            tar.extractall(path, members, numeric_owner=numeric_owner) 
            
        
        safe_extract(tar_handle, path=target_dir)


def read_openpose_surrey_format(filepath: str, fps: int, width: int, height: int) -> Pose:
    """
    Read files of the form "focusnews.071.openpose.tar.xz".
    Assumes a 135 keypoint Openpose model.

    :param filepath:
    :param fps:
    :return:
    """
    with tempfile.TemporaryDirectory(prefix="extract_pose_file") as tmpdir_name:
        # extract tar.xz
        extract_tar_xz_file(filepath=filepath, target_dir=tmpdir_name)

        openpose_dir = os.path.join(tmpdir_name, "openpose")

        # load directory
        poses = load_openpose_135_directory(directory=openpose_dir, fps=fps, width=width, height=height)

    return poses


def formatted_holistic_pose(width=1000, height=1000):

    try:
        import mediapipe as mp
        from pose_format.utils.holistic import holistic_components

    except ImportError:
        raise ImportError("Please install mediapipe with: pip install mediapipe")

    mp_holistic = mp.solutions.holistic
    FACEMESH_CONTOURS_POINTS = [str(p) for p in sorted(set([p for p_tup in list(mp_holistic.FACEMESH_CONTOURS) for p in p_tup]))]

    dimensions = PoseHeaderDimensions(width=width, height=height, depth=width)
    header = PoseHeader(version=0.1, dimensions=dimensions, components=holistic_components("XYZC", 10))
    body = NumPyPoseBody(
        fps=10, data=np.zeros(shape=(1, 1, header.total_points(), 3)), confidence=np.zeros(shape=(1, 1, header.total_points()))
    )
    pose = Pose(header, body)
    return pose.get_components(
        ["POSE_LANDMARKS", "FACE_LANDMARKS", "LEFT_HAND_LANDMARKS", "RIGHT_HAND_LANDMARKS"], {"FACE_LANDMARKS": FACEMESH_CONTOURS_POINTS}
    )


def load_mediapipe_directory(directory: str, fps: int, width: int, height: int) -> Pose:
    """

    :param directory:
    :param fps:
    :return:
    """

    frames = load_frames_directory_dict(directory=directory, pattern="(?:^|\D)?(\d+).*?.json")

    def load_mediapipe_frame(frame):
        def load_landmarks(name, num_points: int):
            points = [[float(p) for p in r.split(",")] for r in frame[name]["landmarks"]]
            points = [(ps + [1.0])[:4] for ps in points]  # Add visibility to all points
            if len(points) == 0:
                points = [[0, 0, 0, 0] for _ in range(num_points)]
            return np.array([[x, y, z] for x, y, z, c in points]), np.array([c for x, y, z, c in points])

        face_data, face_confidence = load_landmarks("face_landmarks", 128)
        body_data, body_confidence = load_landmarks("pose_landmarks", 33)
        lh_data, lh_confidence = load_landmarks("left_hand_landmarks", 21)
        rh_data, rh_confidence = load_landmarks("right_hand_landmarks", 21)
        data = np.concatenate([body_data, face_data, lh_data, rh_data])
        conf = np.concatenate([body_confidence, face_confidence, lh_confidence, rh_confidence])
        return data, conf

    def load_mediapipe_frames():
        max_frames = int(max(frames.keys())) + 1
        pose_body_data = np.zeros(shape=(max_frames, 1, 21 + 21 + 33 + 128, 3), dtype=np.float)
        pose_body_conf = np.zeros(shape=(max_frames, 1, 21 + 21 + 33 + 128), dtype=np.float)
        for frame_id, frame in frames.items():
            data, conf = load_mediapipe_frame(frame)
            pose_body_data[frame_id][0] = data
            pose_body_conf[frame_id][0] = conf
        return NumPyPoseBody(data=pose_body_data, confidence=pose_body_conf, fps=fps)

    pose = formatted_holistic_pose(width=width, height=height)

    pose.body = load_mediapipe_frames()

    return pose


def read_mediapipe_surrey_format(filepath: str, fps: int, width: int, height: int) -> Pose:
    """
    Read files of the form "focusnews.103.mediapipe.tar.xz"
    """
    with tempfile.TemporaryDirectory(prefix="extract_pose_file") as tmpdir_name:
        # extract tar.xz
        extract_tar_xz_file(filepath=filepath, target_dir=tmpdir_name)
        poses_dir = os.path.join(tmpdir_name, "poses")
        # load directory
        pose = load_mediapipe_directory(directory=poses_dir, fps=fps, width=width, height=height)
    return pose


def get_video_metadata(filename: str) -> Dict[str, int]:
    """
    Get metadata from mp4 video.

    :param filename:
    :return:
    """
    cap = cv2.VideoCapture(filename)
    fps = int(cap.get(cv2.CAP_PROP_FPS))
    width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    cap.release()

    return {"fps": fps, "width": width, "height": height}


def milliseconds_to_frame_index(milliseconds: int, fps: int) -> int:
    """
    :param miliseconds:
    :param fps:
    :return:
    """
    return int(fps * (milliseconds / 1000))


def convert_srt_time_to_frame(srt_time: datetime.timedelta, fps: int) -> int:
    """
    datetime.timedelta(seconds=4, microseconds=71000)

    :param srt_time:
    :param fps:
    :return:
    """
    seconds, microseconds = srt_time.seconds, srt_time.microseconds

    milliseconds = int((seconds * 1000) + (microseconds / 1000))

    return milliseconds_to_frame_index(milliseconds=milliseconds, fps=fps)


def reduce_pose_people(pose: Pose):
    pose.body.data = pose.body.data[:, 0:1, :, :]
    pose.body.confidence = pose.body.confidence[:, 0:1, :]



import json

import tensorflow_datasets as tfds
from dotenv import load_dotenv
import hashlib

# noinspection PyUnresolvedReferences
import sign_language_datasets.datasets
from sign_language_datasets.datasets.config import SignDatasetConfig

load_dotenv()

IANA_TAGS = {
    "ch-de": "sgg",
    "ch-fr": "ssr",
    "ch-it": "slf",
}


def signsuisse():
    config = SignDatasetConfig(name="firestore-4", version="1.0.0", include_video=False)
    dataset = tfds.load(name='sign_suisse', builder_kwargs={"config": config})

    # signbank = tfds.load(name='sign_bank', builder_kwargs={"config": config})

    for datum in dataset["train"]:
        uid_raw = "signsuisse" + datum['id'].numpy().decode('utf-8')

        signed_language_original = datum['signedLanguage'].numpy().decode('utf-8')
        signed_language = IANA_TAGS[signed_language_original]
        category = datum['category'].numpy().decode('utf-8')
        url = datum['url'].numpy().decode('utf-8')
        paraphrase = datum['paraphrase'].numpy().decode('utf-8')
        definition = datum['definition'].numpy().decode('utf-8')

        example_text = datum['exampleText'].numpy().decode('utf-8')
        videos = [{
            "id_suffix": "isolated",
            "text": datum['name'].numpy().decode('utf-8'),
            "url": datum['video'].numpy().decode('utf-8'),
            "example": example_text if len(example_text) > 0 else None
        }]

        example_video_url = datum['exampleVideo'].numpy().decode('utf-8')
        if len(example_text) > 0:
            videos.append({
                "id_suffix": "example",
                "text": example_text,
                "url": example_video_url,
                "example": None
            })

        spoken_language = datum['spokenLanguage'].numpy().decode('utf-8')

        id_func = lambda suffix: 'ss' + hashlib.md5((uid_raw+suffix).encode()).hexdigest()
        ids = [id_func(video["id_suffix"]) for video in videos]

        for video in videos:
            if video["text"].lower() == "mit" or video["example"] == "Kaffee mit Zucker und Milch.":
                print("FOUND", video)

            this_id = id_func(video["id_suffix"])
            doc = {
                "uid": this_id,
                "url": video["url"],
                "meta": {
                    "visibility": "unlisted",
                    "name": "SignSuisse: " + video["text"],
                    "language": signed_language,
                    "userId": "ndHslReGkGSJMHxAoSkBeLfjbfU2"
                },
                "related": [_id for _id in ids if _id != this_id],
                "context": {
                    "id": uid_raw,
                    "category": category,
                    "url": url,
                    "example": video["example"],
                    "paraphrase": paraphrase,
                    "definition": definition
                }
            }

            captions = [
                {"language": spoken_language, "transcription": video["text"]},
            ]

            yield doc, captions, None


if __name__ == '__main__':
    with open('signsuisse.jsonl', 'w') as f:
        for doc, captions, _ in signsuisse():
            f.write(json.dumps({"doc": doc, "captions": captions}))
            f.write('\n')



import os

from pose_format import Pose

p = "/home/nlp/amit/sign-language/sign-language-datasets/old/autsl/holistic/train/"
f_path = p + os.listdir(p)[0]

f = open(f_path, "rb").read()
pose = Pose.read(f)

with open("../datasets/autsl/pose.poseheader", "wb") as fw:
    pose.header.write(fw)



def download_youtube(url, dst_path):
    import youtube_dl  # Required for YouTube downloads

    ydl_opts = {"format": "bestvideo", "outtmpl": dst_path}
    with youtube_dl.YoutubeDL(ydl_opts) as ydl:
        try:
            ydl.download([url])
        except youtube_dl.utils.DownloadError as e:
            print("Problem downloading youtube video", url)



import itertools
from os import path

import tensorflow_datasets as tfds
from datasets import tqdm
from dotenv import load_dotenv

# noinspection PyUnresolvedReferences
import sign_language_datasets.datasets

# noinspection PyUnresolvedReferences
import sign_language_datasets.datasets.dgs_corpus
from sign_language_datasets.datasets.config import SignDatasetConfig

from pose_format.utils.holistic import load_holistic

import cv2


def load_video(cap: cv2.VideoCapture):
    while True:
        ret, frame = cap.read()
        if not ret:
            break
        yield cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)


config = SignDatasetConfig(name="full2", version="1.0.0", include_video=True, process_video=False, include_pose=None)
dgs_types = tfds.load(name='dgs_types', builder_kwargs={"config": config})

decode_str = lambda s: s.numpy().decode('utf-8')
for datum in tqdm(dgs_types["train"]):
    _id = decode_str(datum['id'])
    for view, video in zip(datum["views"]["name"], datum["views"]["video"]):
        view = decode_str(view)
        video = decode_str(video)

        pose_file_path = path.join("/home/nlp/amit/WWW/datasets/poses/holistic/dgs_types", f"{_id}_{view}.pose")
        if not path.exists(pose_file_path):
            cap = cv2.VideoCapture(video)
            width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
            height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
            fps = int(cap.get(cv2.CAP_PROP_FPS))
            all_frames = list(load_video(cap))

            pose = load_holistic(all_frames, fps=fps, width=width, height=height, depth=width, progress=False)
            with open(pose_file_path, "wb") as f:
                pose.write(f)

            cap.release()



"""
This SignWriting OCR code is adapted from
https://colab.research.google.com/drive/1x0OupzZNkQW1rCiDjEe1LX5V9k8_YF69#scrollTo=_1YX_pILnjFe
"""
import functools
from typing import List

import numpy as np
from numpy.lib.stride_tricks import as_strided
import cv2
import os


def strided_convolution(image, weight, stride=1):
    im_h, im_w = image.shape
    f_h, f_w = weight.shape

    out_shape = (1 + (im_h - f_h) // stride, 1 + (im_w - f_w) // stride, f_h, f_w)
    out_strides = (image.strides[0] * stride, image.strides[1] * stride, image.strides[0], image.strides[1])
    windows = as_strided(image, shape=out_shape, strides=out_strides)

    return np.tensordot(windows, weight, axes=((2, 3), (0, 1)))


def rgb2bin(img, neg=0.1):
    return np.where(img[:, :, 0] < 200, 1, neg)


def key2swu(key):
    return chr(0x40001 + ((int(key[1:4], 16) - 256) * 96) + ((int(key[4:5], 16)) * 16) + int(key[5:6], 16))


def shape_pos(shape):
    x, y = shape[:2]
    return f"{x:03}x{y:03}"


def crop_whitespace(img):
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    gray = 255 * (gray < 128).astype(np.uint8)  # To invert the text to white
    coords = cv2.findNonZero(gray)  # Find all non-zero points (text)
    x, y, w, h = cv2.boundingRect(coords)  # Find minimum spanning bounding box
    if y > 0:
        y -= 1
        h += 1
    if x > 0:
        x -= 1
        w += 1
    if y + h < img.shape[0]:
        h += 1
    if x + w < img.shape[1]:
        w += 1

    return cv2.copyMakeBorder(img[y : y + h, x : x + w], 2, 2, 2, 2, cv2.BORDER_CONSTANT, value=(255, 255, 255))


@functools.lru_cache()
def get_font():
    from PIL import ImageFont

    dirname = os.path.dirname(__file__)
    font_path = os.path.join(dirname, "assets/SuttonSignWritingOneD.ttf")

    return ImageFont.truetype(font_path, 30)


def image_to_fsw(image: np.ndarray, symbols: List[str]) -> str:
    try:
        from PIL import Image, ImageDraw
    except ImportError:
        raise ImportError("Please install pillow with: pip install Pillow")


    font = get_font()

    # Adding border for conv calc to go over borders
    img_rgb = cv2.copyMakeBorder(image.copy(), 30, 30, 30, 30, cv2.BORDER_CONSTANT, value=(255, 255, 255))
    img_bin = rgb2bin(img_rgb, -0.5)

    img_width, img_height, _ = img_rgb.shape
    final_sign = "M" + shape_pos([500, 500])
    # final_sign = "A" + "".join(symbols) + "M" + shape_pos(img_rgb.shape)

    # Create all convolutions
    templates = []
    convs = []
    for symbol in symbols:
        swu = key2swu(symbol)
        (width, height), (offset_x, offset_y) = font.font.getsize(swu)
        img = Image.new("RGB", (width - offset_x, height - offset_y), (255, 255, 255))
        d = ImageDraw.Draw(img)
        d.text((-offset_x, -offset_y), swu, font=font, fill=(0, 0, 0))

        template = crop_whitespace(np.array(img))
        templates.append(template)

        conv = strided_convolution(img_bin, rgb2bin(template, -0.5))
        convs.append(conv)

    # Save video?
    # out = cv2.VideoWriter('project.avi',cv2.VideoWriter_fourcc(*'DIVX'), 15, img_rgb.shape[:2])

    # Find max symbol
    while len(symbols) > 0:
        # 1. Find best match
        best_match = {"symbol": "", "point": (), "score": 0}
        for symbol, conv in zip(symbols, convs):
            point = np.unravel_index(np.argmax(conv, axis=None), conv.shape)
            score = conv[point]
            if score > best_match["score"]:
                best_match = {"symbol": symbol, "point": point, "score": score}

        idx = symbols.index(best_match["symbol"])

        # 2. Select best match
        pt = best_match["point"][::-1]
        position = [int(500 - img_width / 2) + pt[0], int(500 - img_height / 2) + pt[1]]
        final_sign += best_match["symbol"] + shape_pos(position)

        w, h = templates[idx].shape[:-1]
        cv2.rectangle(img_rgb, pt, (pt[0] + h, pt[1] + w), (0, 0, 255), 1)

        # 3. Remove symbol from list
        # print(best_match)
        # cv2_imshow(templates[idx])
        # cv2_imshow(convs[idx])
        # out.write(img_rgb)
        symbols.pop(idx)
        convs.pop(idx)
        templates.pop(idx)

        # 4. Prevent collisions with this match
        for conv in convs:
            x, y = pt
            conv[y - 2 : y + 4, x - 2 : x + 4] = 0

    return final_sign


if __name__ == "__main__":
    img_rgb = cv2.imread("assets/sign.png")
    symbols = ["S1f520", "S1f528", "S23c04", "S23c1c", "S2fb04", "S2ff00", "S33b10"]

    fsw = image_to_fsw(img_rgb, symbols)
    print("fsw", fsw)
